'''Holds a dataset.'''

# Copyright (c) 2025 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from enum import Enum
import copy
from typing import Any, Dict, List, Optional

import numpy as np
import pandas as pd


from ..config_components.impl.config_component import ConfigComponent
from ..problem_def.task import DataType, TaskType
from ..problem_def.problem_def_interface import ProblemDefInterface
from ..tables.impl.table import Table, TableFactory


def df_to_dict_json_friendly(df: pd.DataFrame) -> Dict:
    '''Same as DataFrame.to_dict(orient='list') but converts NaN to None.'''
    retval = df.replace([np.nan], [None]).to_dict(orient='list')
    return retval


class Error(BaseException):
    '''Base class for all model-related exceptions.'''
    def __str__(self) -> str:
        return self.__repr__()

    def __repr__(self) -> str:
        return self.__class__.__name__ + '(' + ',\n'.join(self.args) + ')'


class DatasetKeyError(Error, KeyError):
    '''Raise when a fit() or predict() method is passed an input dataset
    that is missing required keys.'''


class DatasetValueError(Error, ValueError):
    '''Raise when a dataset has the correct keys but invalid value(s)'''


class MetadataError(Error):
    '''Raise when a dataset is invalid metadata.'''


class RoleName(Enum):
    '''Possible column roles'''
    INDEX = 'index'
    TARGET = 'target'
    ATTRIBUTE = 'attribute'

    # For testing only
    TEST_ROLE = 'test_role'

    # For time series only
    TIME = 'time'
    TIMESERIES_ID = 'timeseries_id'
    PAST_EXOGENOUS = 'past_exogenous'
    FUTURE_EXOGENOUS = 'future_exogenous'


class Column():
    '''Defines a column in a dataset'''
    _name: str

    def __init__(self,
                 name: str):
        self._name = name

    @property
    def name(self) -> str:
        '''The column's name'''
        return str(self._name)

    def __eq__(self, other: Any):
        if not isinstance(other, Column):
            return False
        return self.name == other.name

    def __hash__(self) -> int:
        return self._name.__hash__()

    def __str__(self) -> str:
        return f'Column({self.name})'

    def __repr__(self) -> str:
        return f'ngautonml.wrangler.dataset.Column ({self.name})'


class DatasetKeys(Enum):
    '''Standard keys used in the Dataset object passed between models'''
    DATAFRAME = 'dataframe'
    TARGET = 'target'
    COVARIATES = 'covariates'
    PREDICTIONS = 'predictions'
    PROBABILITIES = 'probabilities'
    HYPERPARAMS = 'hyperparams'
    GROUND_TRUTH = 'ground_truth'

    # Indicator of an error during fitting a pipeline
    ERROR = 'error'

    # Temporary key for migrating to Tables
    DATAFRAME_TABLE = 'dataframe_table'

    # for time series forecasting only
    STATIC_EXOGENOUS = 'static_exogenous'
    DYNAMIC_EXOGENOUS = 'dynamic_exogenous'

    # for keras image processing only
    KERAS_DS = 'keras_ds'
    KERAS_VALIDATE = 'keras_validate'


class Metadata():
    '''Metadata.'''
    _roles: Dict[RoleName, List[Column]]
    _pos_labels: Dict[RoleName, Any]
    _task: Optional[TaskType]
    _data_type: Optional[DataType]
    _problem_def: Optional[ProblemDefInterface]

    def __init__(self,
                 problem_def: Optional[ProblemDefInterface] = None,
                 roles: Optional[Dict[RoleName, List[Column]]] = None,
                 pos_labels: Optional[Dict[RoleName, Any]] = None,
                 task: Optional[TaskType] = None,
                 data_type: Optional[DataType] = None) -> None:
        self._problem_def = problem_def
        self._roles = roles or {}
        self._pos_labels = pos_labels or {}
        if problem_def is None:
            self._task = task
            self._data_type = data_type
        else:
            self._task = task or problem_def.task.task_type
            self._data_type = data_type or problem_def.task.data_type

    def get_conf(self, config_name: str) -> ConfigComponent:
        '''Get a plugin conf ConfigComponent'''
        assert self._problem_def is not None, 'BUG: _problem_def should be resolved in __init__().'
        return self._problem_def.get_conf(config_name=config_name)

    @property
    def roles(self) -> Dict[RoleName, List[Column]]:
        '''roles maps role names to lists of columns with that role.'''
        return self._roles.copy()

    @property
    def task(self) -> Optional[TaskType]:
        '''Data science task that is being approached with this dataset.'''
        return self._task

    @property
    def data_type(self) -> Optional[DataType]:
        '''Type of data contained in this dataset (image, tabular, etc)'''
        return self._data_type

    @property
    def target(self) -> Optional[Column]:
        '''Get name and index of target column if it exists.'''
        if RoleName.TARGET not in self._roles or len(self._roles[RoleName.TARGET]) == 0:
            return None
        assert len(self._roles[RoleName.TARGET]) <= 1, 'Must have at most 1 target column'
        return self._roles[RoleName.TARGET][0]

    @property
    def pos_labels(self) -> Dict[RoleName, Any]:
        '''maps role names to a positive value for columns with that role'''
        return self._pos_labels.copy()

    def override_roles(self, roles=Dict[RoleName, List[Column]]) -> 'Metadata':
        '''Return a copy of metadata with new roles.'''
        other = copy.deepcopy(self)
        other._roles = roles  # pylint: disable=protected-access
        return other

    @property
    def problem_def(self) -> Optional[ProblemDefInterface]:
        '''Get the problem definition object'''
        return self._problem_def


class Dataset(Dict[str, Any]):
    '''Holds a dataset.
    Allows for arbitrary keys and values as long as the keys are strings.

    Standard keys: covariates, target, dataframe
    '''
    _metadata: Metadata

    def __init__(self,
                 *args,
                 metadata: Optional[Metadata] = None,
                 **kwargs):
        super().__init__(*args, **kwargs)
        if metadata is None:
            metadata = Metadata(roles={}, pos_labels={})
        self._metadata = metadata

    def __str__(self) -> str:
        retval = 'Dataset({'
        for key, val in self.items():
            retval = f'{retval}\'{key}\': {self._str_value(val)}\n'
        retval = retval + '})'
        return retval

    def _str_value(self, value: Any) -> str:
        '''String representation of Dataset values.'''
        if isinstance(value, pd.DataFrame):
            return f'pandas.DataFrame(shape = {value.shape}, head = \n{value.head(5)})'
        return str(value)

    def to_prejson(self) -> Dict[str, Any]:
        '''Represent self as a dict that can be transformed to json.'''
        retval: Dict[str, Any] = {}
        for key, val in self.items():
            retval[key] = self._json_value(val)
        return retval

    def _json_value(self, value: Any) -> Any:
        '''Represent a value as something that can be converted to JSON.'''
        if isinstance(value, pd.DataFrame):
            return df_to_dict_json_friendly(value)
        if isinstance(value, Table):
            return value.as_dict()
        # TODO(Merritt): raise an error?
        return value

    @property
    def roles(self) -> Dict[RoleName, List[Column]]:
        '''Information about column roles in the dataset'''
        return self._metadata.roles

    @property
    def metadata(self) -> Metadata:
        '''Config metadata for use by models'''
        return self._metadata

    @property
    def dataframe(self) -> pd.DataFrame:
        '''Return the dataframe at DatasetKeys.DATAFRAME, or raise an error if it doesn't exist.'''
        if DatasetKeys.DATAFRAME_TABLE.value in self:
            return self.dataframe_table.as_(pd.DataFrame)  # type: ignore[attr-defined]  # pylint: disable=line-too-long

        if DatasetKeys.DATAFRAME.value not in self:
            if DatasetKeys.ERROR.value in self:
                raise DatasetKeyError(
                    f'Attempting to extract key "{DatasetKeys.DATAFRAME.value}" '
                    f'as input, instead found keys {list(self.keys())}. '
                    f'Error: {self[DatasetKeys.ERROR.value]}')
            raise DatasetKeyError(
                f'Attempting to extract key "{DatasetKeys.DATAFRAME.value}" '
                f'as input, instead found keys {list(self.keys())}.')
        df_value = self[DatasetKeys.DATAFRAME.value]
        if not isinstance(df_value, pd.DataFrame):
            raise DatasetValueError(
                f'Input key key {DatasetKeys.DATAFRAME.value} must point to '
                f'a pandas.DataFrame, instead found a {type(df_value)}: '
                f'{df_value}')
        return pd.DataFrame(df_value)

    @dataframe.setter
    def dataframe(self, data: pd.DataFrame) -> None:
        self[DatasetKeys.DATAFRAME.value] = data

    @property
    def dataframe_table(self) -> Table:
        '''Return the dataframe at DatasetKeys.DATAFRAME as a Table.'''
        if DatasetKeys.DATAFRAME_TABLE.value in self:
            return self[DatasetKeys.DATAFRAME_TABLE.value]
        if DatasetKeys.DATAFRAME.value not in self:
            if DatasetKeys.ERROR.value in self:
                raise DatasetKeyError(
                    f'Attempting to extract key "{DatasetKeys.DATAFRAME_TABLE.value}" '
                    f'as input, instead found keys {list(self.keys())}. '
                    f'Error: {self[DatasetKeys.ERROR.value]}')
            raise DatasetKeyError(
                f'Attempting to extract key "{DatasetKeys.DATAFRAME_TABLE.value}" '
                f'as input, instead found keys {list(self.keys())}.')
        return TableFactory(self.dataframe)
        # This method will be modified once Table migration is complete

    @dataframe_table.setter
    def dataframe_table(self, data: Table) -> None:
        '''Set the dataframe at DatasetKeys.DATAFRAME_TABLE.'''
        self[DatasetKeys.DATAFRAME_TABLE.value] = data

    @property
    def target_table(self) -> Table:
        '''Return the dataframe at DatasetKeys.TARGET as a Table.'''

        # The following 2 lines will be deleted once Table migration is complete.
        if DatasetKeys.TARGET.value in self:
            return TableFactory(self[DatasetKeys.TARGET.value])

        if self.metadata.target is not None:
            return TableFactory(self.dataframe_table[self.metadata.target.name])
        raise DatasetKeyError(
            f'Attempting to extract key "{DatasetKeys.TARGET.value}" '
            f'as input, instead found keys {list(self.keys())}.')

    @property
    def covariates_table(self) -> Table:
        '''Return the dataframe at DatasetKeys.COVARIATES as a Table.'''
        # The following 2 lines will be deleted once Table migration is complete.
        if DatasetKeys.COVARIATES.value in self:
            return TableFactory(self[DatasetKeys.COVARIATES.value])

        if self.metadata.target is not None:
            return self.dataframe_table.drop(columns=[self.metadata.target.name])

        return self.dataframe_table

    @property
    def ground_truth(self) -> pd.DataFrame:
        '''Return the dataframe at DatasetKeys.GROUND_TRUTH

        or raise an error if it doesn't exist.'''

        if DatasetKeys.GROUND_TRUTH.value not in self:
            if DatasetKeys.ERROR.value in self:
                raise DatasetKeyError(
                    f'Attempting to extract key "{DatasetKeys.GROUND_TRUTH.value}" '
                    f'as input, instead found keys {list(self.keys())}. '
                    f'Error: {self[DatasetKeys.ERROR.value]}')
            raise DatasetKeyError(
                f'Attempting to extract key "{DatasetKeys.GROUND_TRUTH.value}" '
                f'as input, instead found keys {list(self.keys())}.')
        df_value = self[DatasetKeys.GROUND_TRUTH.value]
        if not isinstance(df_value, pd.DataFrame):
            raise DatasetValueError(
                f'Input key key {DatasetKeys.GROUND_TRUTH.value} must point to '
                f'a pandas.DataFrame, instead found a {type(df_value)}: '
                f'{df_value}')
        return pd.DataFrame(df_value)

    @ground_truth.setter
    def ground_truth(self, data: pd.DataFrame) -> None:
        self[DatasetKeys.GROUND_TRUTH.value] = data

    @property
    def predictions(self) -> pd.DataFrame:
        '''Return the dataframe at DatasetKeys.PREDICTIONS
        or raise an error if it doesn't exist.'''

        if DatasetKeys.PREDICTIONS.value not in self:
            if DatasetKeys.ERROR.value in self:
                raise DatasetKeyError(
                    f'Attempting to extract key "{DatasetKeys.PREDICTIONS.value}" '
                    f'as input, instead found keys {list(self.keys())}. '
                    f'Error: {self[DatasetKeys.ERROR.value]}')
            raise DatasetKeyError(
                f'Attempting to extract key "{DatasetKeys.PREDICTIONS.value}" '
                f'as input, instead found keys {list(self.keys())}.')
        df_value = self[DatasetKeys.PREDICTIONS.value]
        if not isinstance(df_value, pd.DataFrame):
            raise DatasetValueError(
                f'Input key key {DatasetKeys.PREDICTIONS.value} must point to '
                f'a pandas.DataFrame, instead found a {type(df_value)}: '
                f'{df_value}')
        return pd.DataFrame(df_value)

    @predictions.setter
    def predictions(self, data: pd.DataFrame) -> None:
        self[DatasetKeys.PREDICTIONS.value] = data

    # TODO(Piggy): There are 20 references to this that lack the name of the target column.
    # This is arguably a bug, but it doesn't matter until we have multiple target columns.
    # The format should be something like this:
    # {
    #    'probabilities': {
    #       'target_column_name_1': [
    #           {value1: 0.6, value2: 0.4, value3: 0.0},
    #           {value1: 0.2, value2: 0.6, value3: 0.2},
    #           {value1: 0.1, value2: 0.2, value3: 0.7},
    #       },
    #       'target_column_name_2': [
    #           {4: 0.6, 5: 0.4, 6: 0.0},
    #           {4: 0.1, 5: 0.8, 6: 0.1},
    #           {4: 0.1, 5: 0.2, 6: 0.7},
    #       },
    #    },
    #    'predictions': {'target_column_name_1': [value1, value2, value3],
    #                    'target_column_name_2': [4, 5, 6]}

    def _check_for_error(self, key: DatasetKeys) -> None:
        '''Check for an error key and raise an error if it exists.'''

        if key.value not in self:
            if DatasetKeys.ERROR.value in self:
                raise DatasetKeyError(
                    f'Attempting to extract key "{key.value}" '
                    f'as input, instead found keys {list(self.keys())}. '
                    f'Error: {self[DatasetKeys.ERROR.value]}')
            raise DatasetKeyError(
                f'Attempting to extract key "{key.value}" '
                f'as input, instead found keys {list(self.keys())}.')

    @property
    def probabilities(self) -> Table:
        '''
        Return the table at `DatasetKeys.PROBABILITIES`
        or raise an error if it doesn't exist.
        '''
        self._check_for_error(DatasetKeys.PROBABILITIES)

        table = self[DatasetKeys.PROBABILITIES.value]

        if not isinstance(table, Table):
            raise DatasetValueError(
                f'Input key key {DatasetKeys.PROBABILITIES.value} must point to '
                f'a Table, instead found a {type(table)}: '
                f'{table}'
            )

        return table

    @probabilities.setter
    def probabilities(self, data: Table) -> None:
        self[DatasetKeys.PROBABILITIES.value] = data

    def output(self, override_metadata: Optional[Metadata] = None) -> 'Dataset':
        '''Return a new dataset containing same metadata and nothing else.

        Metadata can also be overriden, all preexisting metadata will be lost.
        Generally called to create a new dataset to fill with a model's output.
        '''
        return self.__class__(metadata=override_metadata or self._metadata)

    def get_dataframe(self) -> pd.DataFrame:
        '''Return the dataframe at DatasetKeys.DATAFRAME, or raise an error if it doesn't exist.'''

        return self.dataframe

    def sorted_columns(self) -> 'Dataset':
        '''Sort columns of a dataset.

        This sorts DATAFRAME and COVARIATES by column name.

        This is needed so that the distributed algorithms operate
        on the same columns in the same order.
        '''
        # TODO(Piggy/Dan): If this becomes a performance problem
        #   we can cache the column order and only sort if the
        #   columns change.
        retval = self.output()
        retval.update(self)

        keys: List[str] = []
        if DatasetKeys.DATAFRAME.value in self:
            keys.append(DatasetKeys.DATAFRAME.value)

        if DatasetKeys.COVARIATES.value in self:
            keys.append(DatasetKeys.COVARIATES.value)

        for key in keys:
            df = self[key]
            if not isinstance(df, pd.DataFrame):
                raise DatasetValueError(
                    f'Expected dataframe at {key}, instead found {df} of type {type(df)}.'
                )
            df.columns = df.columns.astype(str)
            # ^ if column names have mixed typing they cannot be sorted
            retval[key] = (
                df.sort_index(axis=1))
        return retval



'''Autoloader for the template catalog.'''

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from pathlib import Path

from ...catalog.plugin_catalog import PluginCatalog
from .pipeline_template import PipelineTemplate
from .template import TemplateCatalog


class TemplateCatalogAuto(PluginCatalog[PipelineTemplate], TemplateCatalog):
    '''Template catalog that automatically loads all templates from a directory.

    Templates are in ../templates.
    '''

    def __init__(self, **kwargs):
        super().__init__('templates', default_root=Path(__file__).parents[2], **kwargs)



'''An object that glues together a sequence of catalogs.'''

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from typing import Any, Dict, Iterable, List, Optional, Set, Tuple, TypeVar, Union

from .catalog import Catalog, CatalogLookupError
from .memory_catalog import MemoryCatalog

T = TypeVar('T')


class CatalogShelf(Catalog[T]):
    '''Catalog object that holds a stack of other Catalog objects.'''
    _catalogs: List[Catalog[T]]
    _kwargs: Dict[str, Any]

    # pylint: disable=super-init-not-called
    def __init__(self, *catalogs: Catalog[T], **kwargs):
        self._catalogs = list(catalogs)
        self._kwargs = kwargs

    @property
    def tagtypes(self) -> Set[str]:
        '''Retrieve list of all tag types present in catalog'''
        retval = set()
        for catalog in self._catalogs:
            retval.update(catalog.tagtypes)
        return retval

    def all_objects(self) -> Iterable[T]:
        '''Retrieve all objects in all catalogs.

        Note that some objects may be returned which are otherwise
        masked during lookups.
        '''
        for catalog in self._catalogs:
            yield from catalog.all_objects()

    def items(self) -> Iterable[Tuple[str, T]]:
        for catalog in self._catalogs:
            for key, obj in catalog.items():
                yield (key, obj)

    def tagvals(self, tagtype: str) -> Set[str]:
        '''Retrieve all tag values present for a given tag type'''
        retval = set()
        for catalog in self._catalogs:
            retval.update(catalog.tagvals(tagtype=tagtype))
        return retval

    def lookup_by_name(self, name: str) -> Any:
        '''Find an object registered under name.

        The object comes from the first catalog with the name.
        '''
        for catalog in self._catalogs:
            try:
                return catalog.lookup_by_name(name=name)
            except CatalogLookupError:
                pass
        raise CatalogLookupError(name)

    def lookup_by_tag_and(self, **tags: Union[str, Iterable[str]]) -> Dict[str, T]:
        '''Find objects with a certain tag type and tag.'''
        result: Dict[str, T] = {}
        for catalog in reversed(self._catalogs):
            result.update(catalog.lookup_by_tag_and(**tags))
        return result

    def prepend(self, catalog: Catalog[T]) -> None:
        '''Add catalog to the front of the list of catalogs.'''
        catalogs = [catalog] + self._catalogs
        self._catalogs = catalogs

    def register(self, obj: T, name: Optional[str] = None,
                 tags: Optional[Dict[str, Union[str, List[str]]]] = None) -> str:
        '''Register a new object with the first catalog.'''
        assert isinstance(self._catalogs[0], MemoryCatalog), (
            'You can not register on a shelf unless the first catalog is a MemoryCatalog.')
        first: MemoryCatalog = self._catalogs[0]
        return first.register(obj=obj, name=name, tags=tags)



'''Load new data from a queue.'''

from queue import Empty, Queue
from typing import Dict, List, Optional

from ..config_components.dataset_config import DatasetConfig
from ..config_components.impl.config_component import InvalidValueError
from ..wrangler.dataset import Dataset

from .impl.data_loader_catalog import DataLoaderCatalog
from .memory_data_loader import MemoryDataLoader


class QueueLoader(MemoryDataLoader):
    '''Load poll data from a queue.'''

    name: str = 'queue'
    tags: Dict[str, List[str]] = {
        'input_format': ['pandas_dataframe'],
        'loaded_format': ['pandas_dataframe'],
    }

    _poll_queue: Optional[Queue] = None
    _timeout: Optional[float] = None

    def __init__(self, config: DatasetConfig):
        '''Create a new QueueLoader.

        Args:
            config: The configuration for the loader.
        '''
        if 'poll_queue' in config.params:
            poll_queue_var_name = config.params['poll_queue']
            self._poll_queue = self._lookup_var(poll_queue_var_name)
            if not isinstance(self._poll_queue, Queue):
                raise InvalidValueError(
                    f'Expecting local variable for poll {poll_queue_var_name} to be '
                    f'a queue.Queue; instead found {type(self._poll_queue)}'
                )
        if 'poll_timeout' in config.params:
            self._timeout = config.params['poll_timeout']
        super().__init__(config=config)

    def _poll(self, timeout: Optional[float] = None) -> Optional[Dataset]:
        '''Poll the queue for data.

        Args:
            timeout: The time to wait for data. If it is None, we
                use the configured timeout for the loader. If that is None,
                we wait forever.

        Returns:
            The data from the queue, or None if the queue times out.
        '''
        if self._poll_queue is None:
            return None
        timeout = timeout if timeout is not None else self._timeout
        try:
            return self._poll_queue.get(timeout=timeout)
        except Empty:
            return None


def register(catalog: DataLoaderCatalog) -> None:
    '''Register all the objects in this file with the catalog.

    Args:
        catalog: The catalog to register the loader with.
    '''
    catalog.register(QueueLoader)



'''Tests from ExtractColumnsByRoleModel'''

# Copyright (c) 2024 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

# pylint: disable=missing-function-docstring, duplicate-code

from pathlib import Path
from typing import Dict, Any, List, Optional

import pandas as pd

from .extract_columns_by_role import ExtractColumnsByRoleModel, ExtractColumnsByRoleModelInstance
from ..config_components.dataset_config import DatasetConfig
from ..data_loaders.local_data_loader import LocalDataLoader
from ..wrangler.dataset import Dataset, RoleName, Metadata, Column


def valid_csv() -> str:
    '''Returns a path (in the form of a string) to a valid csv file.'''
    module_parent = Path(__file__).parents[2]
    path = module_parent / 'examples' / 'classification' / 'credit.csv'
    return str(path)


def make_dataset_clause() -> Dict[str, Any]:
    retval = {
        'config': 'local',
        'train_path': valid_csv(),
        'column_roles': {
            'target': {
                'name': 'class'
            },
            'index': {
                'name': 'checking_status'
            }
        }
    }
    return retval


def test_simple_extraction() -> None:
    clause: Dict[str, Any] = make_dataset_clause()
    dataset_config = DatasetConfig(clause=clause)
    data_loader = LocalDataLoader(dataset_config)
    dataset = data_loader.load_train()
    desired_role: RoleName = RoleName.TARGET

    dut: ExtractColumnsByRoleModelInstance = ExtractColumnsByRoleModel().instantiate(
        desired_roles=desired_role
    )
    dut_result: Optional[Dataset] = dut.predict(dataset=dataset)
    assert dut_result is not None
    result_cols: List[str] = list(dut_result.dataframe.columns)
    assert len(result_cols) == 1
    assert result_cols[0] == 'class'


def test_nonexistent_role_extraction() -> None:
    desired_roles: List[RoleName] = [RoleName.TIME]

    fake_df = pd.DataFrame({
        'a': [1, 2], 'b': [3, 4]
    })
    fake_dataset = Dataset(
        dataframe=fake_df,
        metadata=Metadata(
            roles={
                RoleName.TARGET: [Column('a')]
            })
    )

    dut: ExtractColumnsByRoleModelInstance = ExtractColumnsByRoleModel().instantiate(
        desired_roles=desired_roles
    )
    got = dut.predict(dataset=fake_dataset)
    assert got is not None
    assert got['dataframe'].shape == (0, 0)


def test_multi_role_extraction() -> None:
    clause: Dict[str, Any] = make_dataset_clause()
    dataset_config = DatasetConfig(clause=clause)
    data_loader = LocalDataLoader(dataset_config)
    dataset = data_loader.load_train()
    desired_roles: List[RoleName] = [RoleName.TARGET, RoleName.INDEX]

    dut: ExtractColumnsByRoleModelInstance = ExtractColumnsByRoleModel().instantiate(
        desired_roles=desired_roles
    )
    dut_result: Optional[Dataset] = dut.predict(dataset=dataset)
    assert dut_result is not None
    result_cols: List[str] = sorted(list(dut_result.dataframe.columns))
    assert len(result_cols) == 2
    assert result_cols[0] == 'checking_status'
    assert result_cols[1] == 'class'


def test_multi_role_extraction_with_one_nonexistent() -> None:
    clause: Dict[str, Any] = make_dataset_clause()
    dataset_config = DatasetConfig(clause=clause)
    data_loader = LocalDataLoader(dataset_config)
    dataset = data_loader.load_train()
    desired_roles: List[RoleName] = [RoleName.TARGET, RoleName.TIME]

    dut: ExtractColumnsByRoleModelInstance = ExtractColumnsByRoleModel().instantiate(
        desired_roles=desired_roles
    )
    dut_result: Optional[Dataset] = dut.predict(dataset=dataset)
    assert dut_result is not None
    result_cols: List[str] = list(dut_result.dataframe.columns)
    assert len(result_cols) == 1
    assert result_cols[0] == 'class'



'''Tests for local_data_loader.py'''
from pathlib import Path
from typing import Any, Dict, Optional

# Copyright (c) 2024 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

import pandas as pd

from ..config_components.dataset_config import DatasetConfig
from ..problem_def.problem_def import ProblemDefinition
from ..problem_def.task import TaskType
from ..wrangler.dataset import DatasetKeys, RoleName
from .local_data_loader import LocalDataLoader


# pylint: disable=missing-function-docstring,duplicate-code


def valid_data_path(filename: Optional[str] = None) -> str:
    '''Returns a path (in the form of a string) to a valid csv file.'''
    module_parent = Path(__file__).parents[2]
    if filename is None:
        filename = 'credit.csv'
    path = module_parent / 'examples' / 'classification' / filename
    return str(path)


def make_dataset(train_path: str) -> Dict[str, Any]:
    retval = {
        'config': 'local',
        'train_path': train_path,
        'column_roles': {
            'target': {
                'name': 'class',
                'pos_label': 'good'
            }
        },
    }
    return retval


def test_load_simple_dataset() -> None:
    clause = make_dataset(valid_data_path())
    config = DatasetConfig(clause=clause)
    dut = LocalDataLoader(config=config)
    result = dut.load_train()

    assert result is not None
    train_df = result.dataframe
    assert train_df.shape == (1000, 21)
    assert train_df.columns[0] == 'checking_status'
    assert train_df.columns[20] == 'class'


def test_no_target() -> None:
    clause = {
        'config': 'local',
        'train_path': valid_data_path(),
        'column_roles': {
            'index': {
                'id': 0
            }
        }
    }
    config = DatasetConfig(clause=clause)
    dut = LocalDataLoader(config=config)
    dataset = dut.load_train()
    assert dataset is not None
    assert dataset.metadata.target is None


def test_arff() -> None:
    clause = {
        'config': 'local',
        'train_path': valid_data_path(filename='dataset_31_credit-g.arff'),
        'column_roles': {
            'target': {
                'name': 'class'
            }
        }
    }
    config = DatasetConfig(clause=clause)
    dut = LocalDataLoader(config=config)
    got = dut.load_train()
    assert got is not None
    target = got.metadata.target
    assert target is not None
    assert 'class' == target.name

    assert DatasetKeys.DATAFRAME.value in got

    train_df = got.dataframe
    assert isinstance(train_df, pd.DataFrame)
    assert train_df.shape == (1000, 21)


def test_load_test() -> None:
    clause = {
        'config': 'simple',
        'train_path': valid_data_path(filename='credit-train.csv'),
        'test_path': valid_data_path(filename='credit-test.csv'),
        'column_roles': {
            'target': {
                'name': 'class'
            }
        },
    }
    config = DatasetConfig(clause=clause)
    dut = LocalDataLoader(config=config)
    result = dut.load_test()
    assert result is not None
    test_df = result.dataframe
    assert test_df.shape == (200, 20)
    assert 'class' not in test_df.columns
    assert 'own_telephone' in test_df.columns


def test_load_nonexistent_testdata() -> None:
    '''If there is no test data, load_test() returns None'''
    clause = make_dataset(valid_data_path())
    config = DatasetConfig(clause=clause)
    dut = LocalDataLoader(config=config)
    result = dut.load_test()
    assert result is None


CLASSIFICATION_TASK_CLAUSE = {
    'problem_type': {
        'task': 'binary_classification'
    },
    'dataset': {
        'config': 'ignore'
    }
}


def test_task_in_metadata_with_load_train() -> None:
    clause = make_dataset(valid_data_path())
    problem_def_for_task = ProblemDefinition(CLASSIFICATION_TASK_CLAUSE)
    config = DatasetConfig(clause=clause, problem_def=problem_def_for_task)
    dut = LocalDataLoader(config=config)
    result = dut.load_train()

    assert result is not None
    assert result.metadata.task == TaskType.BINARY_CLASSIFICATION


def test_task_in_metadata_with_load_test() -> None:
    clause = {
        'config': 'local',
        'train_path': valid_data_path(filename='credit-train.csv'),
        'test_path': valid_data_path(filename='credit-test.csv'),
        'column_roles': {
            'target': {
                'name': 'class'
            }
        },
    }
    problem_def_for_task = ProblemDefinition(CLASSIFICATION_TASK_CLAUSE)
    config = DatasetConfig(clause=clause, problem_def=problem_def_for_task)
    dut = LocalDataLoader(config=config)
    result = dut.load_test()

    assert result is not None
    assert result.metadata.task == TaskType.BINARY_CLASSIFICATION


def test_load_arbitrary_csv() -> None:
    clause = make_dataset(valid_data_path(filename='credit-train.csv'))
    config = DatasetConfig(clause=clause)
    dut = LocalDataLoader(config=config)

    got = dut.dataset(data=valid_data_path(filename='credit-test.csv'),
                      cols=['class', 'own_telephone'])
    assert got.dataframe.shape == (200, 2)
    assert set(got.dataframe.columns) == {'class', 'own_telephone'}
    assert got.metadata.roles[RoleName.TARGET][0].name == 'class'



'''Object that listens for inbound messages and sends outbound messages'''

# Copyright (c) 2024 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.
import abc
from queue import Queue
from typing import Dict, List, Set

from ...config_components.distributed_config import DistributedConfig
from ...neighbor_manager.node_id import NodeID
from ...wrangler.constants import Defaults


class Communicator(metaclass=abc.ABCMeta):
    '''Object that listens for inbound messages and sends outbound messages'''
    name: str
    tags: Dict[str, List[str]]
    _known_neighbors: Set[NodeID]
    _my_id: NodeID
    _distributed: DistributedConfig

    # All subclasses must implement this __init__ signature.
    # When we get something from a catalog we don't otherwise know what
    # arguments it needs.
    def __init__(self,
                 my_id: NodeID,
                 known_neighbors: Set[NodeID],
                 distributed: DistributedConfig
                 ):
        self._my_id = my_id
        self._known_neighbors = known_neighbors
        self._distributed = distributed

    @abc.abstractmethod
    def start(self,
              queue: Queue,
              timeout: float = Defaults.LISTENER_TIMEOUT
              ) -> None:
        '''Start listening.'''

    @abc.abstractmethod
    def stop(self) -> None:
        '''Stop the listener'''

    @abc.abstractmethod
    def send(self,
             dest_ids: List[NodeID],
             payload: bytes) -> int:
        '''Send a message to the nodes with the given dest_ids.

        Return value is the size of the message sent.
        '''

    @abc.abstractmethod
    def send_all(self, payload: bytes) -> int:
        '''Send a message to all known neighbors.

        Return value is a count of neighbors we sent to.
        '''

    @property
    def known_neighbors(self) -> Set[NodeID]:
        '''Provide a modifiable set of neighbor NodeIDs.

        This is how the communicator tells the discoverer
        about new neighbors.
        '''
        return self._known_neighbors

    @property
    def my_id(self) -> NodeID:
        '''What do I think my own ID is?'''
        return self._my_id



'''Uses a single train/test split and only produces predictions for test split.

Does not do true cross-validation.
'''
# pylint: disable=too-many-arguments,unused-argument

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from typing import Dict

from ..catalog.catalog import upcast
from ..executor.executor import ExecutablePipeline, Executor
from ..generator.bound_pipeline import BoundPipeline
from ..generator.designator import Designator
from ..instantiator.executable_pipeline import PipelineResult
from ..instantiator.instantiator_factory import InstantiatorFactory
from ..splitters.impl.splitter import SplitDataset
from .impl.cross_validator import CrossValidator, CrossValidatorCatalog


class SingleFoldValidator(CrossValidator):
    '''Uses a single train/test split and only produces predictions for test split.'''

    _name = 'single_fold_validator'
    _tags = None

    def validate_pipelines(self,
                           split_dataset: SplitDataset,
                           bound_pipelines: Dict[Designator, BoundPipeline],
                           instantiator: InstantiatorFactory,
                           executor: Executor,
                           **overrides
                           ) -> Dict[Designator, PipelineResult]:
        '''Uses a single train/test split and only produces predictions for test split.'''

        assert len(split_dataset.folds) == 1, (
            f'BUG: SingleFoldValidator got split dataset with {len(split_dataset)} folds.  '
            'Exactly one fold required.'
        )

        executable_pipelines: Dict[Designator, ExecutablePipeline] = {}
        for designation, pipe in bound_pipelines.items():
            instance = instantiator.instantiate(
                kind=executor.kind,
                pipeline=pipe)
            executable_pipelines[designation] = instance

        executor.fit(split_dataset.folds[0].train, executable_pipelines)
        print('Fit all pipelines.')

        train_predictions = executor.predict(
            dataset=split_dataset.folds[0].validate,
            pipelines=executable_pipelines)
        print(f'Got predictions for {len(train_predictions)} pipelines.')

        retval = {
            des: PipelineResult(
                prediction=train_predictions[des].prediction,
                bound_pipeline=pipe,
                split_dataset=split_dataset)
            for des, pipe in bound_pipelines.items()}
        return retval


def register(catalog: CrossValidatorCatalog, *args, **kwargs) -> None:
    '''Register all the objects in this file.'''
    val = SingleFoldValidator()
    catalog.register(val, val.name, upcast(val.tags))



'''Base class for Boolean deciders.'''
# Copyright (c) 2024 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

import abc
from typing import Dict, List, Mapping, Optional

from aenum import Enum as AEnum  # type: ignore[import-untyped]

from ..algorithms.impl.neighbor_metadata_interface import NeighborMetadataInterface
from ..config_components.distributed_config import DeciderConfig, DecidersConfig
from ..neighbor_manager.node_id import NodeID

from .impl.decider import Decider
from .impl.decider_catalog import DeciderCatalog
from .impl.decider_auto import DeciderCatalogAuto


class BooleanDeciderConfig(DeciderConfig, metaclass=abc.ABCMeta):
    '''Configuration for BooleanDecider.'''

    def __init__(self, name: str, clause: dict) -> None:
        super().__init__(name=name, clause=clause)

    def __str__(self) -> str:
        return f'boolean({",".join([str(c) for c in self._clause])})'

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, BooleanDeciderConfig):
            return False
        return all(self._clause[decider] == other._clause[decider] for decider in self._clause)

    class Keys(AEnum):
        '''Keys for the clause.'''
        ENABLED = 'enabled'

    @property
    def subdeciders(self) -> Dict[str, DeciderConfig]:
        '''Return the subdeciders.'''
        return DecidersConfig(self._clause).decider_configs


class BooleanDecider(Decider, metaclass=abc.ABCMeta):
    '''Decider that decides True if all subdeciders decide True.'''
    name = 'boolean'
    tags = {}
    _config: BooleanDeciderConfig
    _subdeciders: Dict[str, Decider]

    def __init__(self,
                 name: Optional[str] = None,
                 tags: Optional[Dict[str, List[str]]] = None,
                 config: Optional[DeciderConfig] = None,
                 catalog: Optional[DeciderCatalog] = None
                 ) -> None:
        super().__init__(name=name, tags=tags, config=config)
        if catalog is None:
            catalog = DeciderCatalogAuto()
        self._catalog = catalog
        self._subdeciders = {}
        for decider, clause in self._config.subdeciders.items():
            subdecider = self._catalog.lookup_by_name(decider)(config=clause)
            self._subdeciders[decider] = subdecider

    def _all_decisions(
            self,
            my_meta: NeighborMetadataInterface,
            neighbor_id: NodeID,
            neighbors: Mapping[NodeID, NeighborMetadataInterface]) -> List[bool]:
        '''All the decisions for this operator.'''
        return [decider.decide(my_meta=my_meta, neighbor_id=neighbor_id, neighbors=neighbors)
                for decider in self._subdeciders.values()]


def register(catalog: DeciderCatalog) -> None:
    '''There are no catalog objects in this file.'''
    _ = catalog



'''Tests for the OutputConfig object'''
from pathlib import Path

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

import pytest

from ..wrangler.constants import FileType
from ..executor.executor_kind import ExecutorKind

from .output_config import OutputConfig, OutputConfigError
# pylint: disable=missing-function-docstring
# pylint: disable=redefined-outer-name,duplicate-code


@pytest.fixture(scope="session")
def tmp_path(tmp_path_factory: pytest.TempPathFactory) -> Path:
    return tmp_path_factory.mktemp("data")


def test_output_config_sunny_day(tmp_path: Path) -> None:
    tmp_dir = tmp_path / "sub"
    tmp_dir.mkdir()
    dut = OutputConfig(clause={
        'path': str(tmp_dir),
        'instantiations': [
            'stub_executor_kind',
            'simple'
        ],
        'file_type': 'csv'
    })
    dut.validate()
    assert dut.path == tmp_dir
    assert dut.instantiations == set([ExecutorKind('simple'), ExecutorKind('stub_executor_kind')])
    assert dut.file_type == FileType.CSV


def test_illegal_instantiation(tmp_path: Path) -> None:
    tmp_dir = tmp_path / "sub"
    dut = OutputConfig(clause={
        'path': str(tmp_dir),
        'instantiations': [
            'stub_executor_kind',
            'fake_instantiation'
        ],
        'file_type': 'csv'
    })
    with pytest.raises(OutputConfigError, match=r'fake_instantiation'):
        dut.validate()


def test_illegal_file_type(tmp_path: Path) -> None:
    tmp_dir = tmp_path / "sub"
    dut = OutputConfig(clause={
        'path': str(tmp_dir),
        'file_type': 'fake_file_type'
    })
    with pytest.raises(OutputConfigError, match=r'fake_file_type'):
        dut.validate()



'''Tests for pipeline_loader_auto.py.'''

# Copyright (c) 2024 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

import os
from pathlib import Path

import pytest

from ...algorithms.impl.algorithm_auto import AlgorithmCatalogAuto
from ...generator.bound_pipeline import BoundPipeline
from ...wrangler.constants import PACKAGE_NAME

from .pipeline_loader import PipelineLoader
from .pipeline_loader_auto import PipelineLoaderCatalogAuto

# pylint: disable=missing-function-docstring,redefined-outer-name,duplicate-code


FAKE_PIPELINE_LOADER_SOURCE = '''
from typing import Any, List, Optional
from ngautonml.generator.bound_pipeline import BoundPipeline
from ngautonml.generator.designator import Designator
from ngautonml.pipeline_loaders.impl.pipeline_loader import PipelineLoader
from ngautonml.pipeline_loaders.impl.pipeline_loader_catalog import PipelineLoaderCatalog

class FakePipelineLoader(PipelineLoader):
    name = 'fake_pipeline_loader'
    tags = {
        'some_tag': ['some_value']
    }

    def _load(self, name) -> BoundPipeline:
        return BoundPipeline(name=Designator(self.name))


def register(catalog: PipelineLoaderCatalog, **kwargs):
    catalog.register(FakePipelineLoader(**kwargs))
'''


@pytest.fixture(scope="session")
def tmp_path(tmp_path_factory: pytest.TempPathFactory) -> Path:
    retval = tmp_path_factory.mktemp("data")
    source_path = (retval / f'.{PACKAGE_NAME}' / 'plugins'
                   / 'pipeline_loaders' / 'fake_pipeline_loader.py')
    source_path.parent.mkdir(parents=True, exist_ok=True)
    with source_path.open('wt') as file:
        file.write(FAKE_PIPELINE_LOADER_SOURCE)
    return retval


def test_pipeline_loader_catalog_auto() -> None:
    dut = PipelineLoaderCatalogAuto(algorithm_catalog=AlgorithmCatalogAuto())
    got = dut.lookup_by_name('just').load('connect')
    assert isinstance(got, BoundPipeline)


def test_home_directory_auto(tmp_path) -> None:
    os.environ['HOME'] = str(tmp_path)
    dut = PipelineLoaderCatalogAuto(algorithm_catalog=AlgorithmCatalogAuto())
    got = dut.lookup_by_name('fake_pipeline_loader')
    assert isinstance(got, PipelineLoader)
    assert got.tags['some_tag'] == ['some_value']



'''Performs k-fold cross-validation.'''

# Copyright (c) 2025 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

# pylint: disable=duplicate-code
from typing import Dict, List, Optional

import pandas as pd


from ..catalog.catalog import upcast
from ..executor.executor import Executor
from ..generator.bound_pipeline import BoundPipeline
from ..generator.designator import Designator
from ..instantiator.executable_pipeline import (ExecutablePipeline,
                                                PipelineResult,
                                                PredictError)
from ..instantiator.instantiator_factory import InstantiatorFactory
from ..splitters.impl.splitter import SplitDataset
from ..tables.impl.table import TableFactory
from ..wrangler.dataset import Dataset, DatasetKeys
from .impl.cross_validator import CrossValidator, CrossValidatorCatalog


class KFoldCrossValidator(CrossValidator):
    '''Standard k-fold cross-validator.

    Works for single-fold validation as well
    '''

    _name = 'k_fold_cross_validator'

    def _concatenate_predictions(self,
                                 datasets: List[Optional[Dataset]]) -> Optional[Dataset]:
        '''Concatenate a list of fold prediction Datasets into a single Dataset
        that covers the entire train set.'''
        assert len(datasets) > 0, 'BUG: _concatenate() called when length of dataset list is 0'

        if any(dset is None for dset in datasets):
            return None

        assert datasets[0] is not None
        retval = datasets[0].output()

        errors: List[PredictError] = []
        for dset in datasets:
            assert dset is not None
            if DatasetKeys.ERROR.value in dset:
                errors.append(dset[DatasetKeys.ERROR.value])

        if len(errors) > 0:
            # handle errors
            error_text = 'Pipeline errors occurred during cross-validation: \n' + '\n'.join([
                str(error) for error in errors
            ])
            retval[DatasetKeys.ERROR.value] = PredictError(error_text)
            return retval

        retval.predictions = datasets[0].predictions
        for dset in datasets[1:]:
            assert dset is not None
            retval.predictions = pd.concat([retval.predictions, dset.predictions], axis=0)

        if DatasetKeys.PROBABILITIES.value in datasets[0]:
            dsets: List[pd.DataFrame] = []
            for dset in datasets:
                assert dset is not None
                dsets.append(dset.probabilities.as_(pd.DataFrame))  # type: ignore[attr-defined]
            probabilities_df = pd.concat(dsets, axis=0)
            retval.probabilities = TableFactory(probabilities_df)

        return retval

    def _concatenate_all(
            self,
            predictions: Dict[Designator,
                              List[Optional[Dataset]]]) -> Dict[Designator, Optional[Dataset]]:
        '''Concatenate each list of fold prediction Datasets into a single Dataset
        that covers the entire train set.'''
        retval: Dict[Designator, Optional[Dataset]] = {
            des: self._concatenate_predictions(data) for des, data in predictions.items()}
        return retval

    def validate_pipelines(self,
                           split_dataset: SplitDataset,
                           bound_pipelines: Dict[Designator, BoundPipeline],
                           instantiator: InstantiatorFactory,
                           executor: Executor,
                           **overrides
                           ) -> Dict[Designator, PipelineResult]:
        '''Do cross-validation by running pipelines on a split dataset.

        Returns the split dataset and an ExecutorResult for each pipeline.'''
        # For each fold, we will append that fold's predictions to the list of datasets for the
        # corresponding pipeline
        train_predictions: Dict[Designator, List[Optional[Dataset]]] = {
            des: [] for des in bound_pipelines}

        executable_pipelines: Dict[Designator, ExecutablePipeline] = {}

        for fold in split_dataset.folds:
            # Instantiate new executable pipelines for each bound pipeline
            # TODO(Piggy/Merritt): run this cross validator in parallel to take advantage of
            #       separate executable pipelines existing for each fold
            executable_pipelines = instantiator.instantiate_all(kind=executor.kind,
                                                                pipelines=bound_pipelines)
            # Fit all executable pipelines on this fold's train set
            executor.fit(dataset=fold.train, pipelines=executable_pipelines)
            # Predict with all executable pipelines on this fold's validate set
            fold_predictions = executor.predict(dataset=fold.validate,
                                                pipelines=executable_pipelines)

            # For each fold and pipeline, append that fold's predictions to that pipeline's
            # train_predictions
            for des, datasets in train_predictions.items():
                datasets.append(fold_predictions[des].prediction)

        # For each pipeline, concatenate its train predictions from a list of datsets into one
        # dataset
        train_predictions_concat = self._concatenate_all(predictions=train_predictions)

        retval: Dict[Designator, PipelineResult] = {
            des: PipelineResult(
                prediction=train_predictions_concat[des],
                bound_pipeline=bound_pipelines[des],
                split_dataset=split_dataset)
            for des in bound_pipelines}
        return retval


def register(catalog: CrossValidatorCatalog, *unused_args, **unused_kwargs) -> None:
    '''Register the cross-validator in the catalog.'''
    val = KFoldCrossValidator()
    catalog.register(val, val.name, upcast(val.tags))



'''Tests for fittable_algorithm_instance.py'''
import pickle
from typing import Optional
import pytest

from ...wrangler.dataset import Dataset

from .algorithm import Algorithm
from .fake_algorithm import FakeAlgorithm
from .fittable_algorithm_instance import (FittableAlgorithmInstance,
                                          UntrainedError,
                                          DeserializationError)

# pylint: disable=missing-class-docstring,missing-function-docstring
# pylint: disable=protected-access,duplicate-code


class FakeFitterInstance(FittableAlgorithmInstance):
    _some_attr: str

    def __init__(self, parent: Algorithm):
        super().__init__(parent=parent)
        self._some_attr = '1'

    def fit(self, dataset: Optional[Dataset]) -> None:
        self._some_attr = '2'
        self._trained = True

    def predict(self, dataset: Optional[Dataset]) -> Optional[Dataset]:
        if dataset is None:
            return None
        dataset['some_attr'] = self._some_attr
        return dataset


def test_serialize_deserialize_round_trip() -> None:
    dut = FakeFitterInstance(parent=FakeAlgorithm())
    dut.fit(dataset=Dataset())
    got = dut.deserialize(dut.serialize())
    assert isinstance(got, FakeFitterInstance)
    assert got._some_attr == '2'


def test_serialize_not_trained() -> None:
    dut = FakeFitterInstance(parent=FakeAlgorithm())
    with pytest.raises(UntrainedError, match='serialize'):
        _ = dut.serialize()


def test_deserialize_not_pickle() -> None:
    dut = FakeFitterInstance(parent=FakeAlgorithm())
    with pytest.raises(DeserializationError, match='pickle'):
        _ = dut.deserialize(b'not a pickle')


def test_deserialize_not_pickled_instance() -> None:
    dut = FakeFitterInstance(parent=FakeAlgorithm())
    with pytest.raises(DeserializationError, match='float'):
        _ = dut.deserialize(pickle.dumps(1.234))



'''Tests for identity model.'''
from .identity import IdentityModel
from ..wrangler.dataset import Dataset
# pylint: disable=missing-function-docstring,duplicate-code

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.


def test_sunny_day() -> None:
    model = IdentityModel().instantiate()

    dataset = Dataset(
        a=[1, 2, 3],
        b='something else'
    )

    got = model.predict(dataset=dataset)
    assert got == dataset



'''Tests for the AlgorithmInstance class implementations'''
from typing import Optional

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from ...wrangler.dataset import Dataset

from .algorithm import Algorithm
from .sklearn_like_algorithm_instance import SklearnLikeAlgorithmInstance

# pylint: disable=missing-function-docstring,duplicate-code
# pylint: disable=missing-class-docstring


class FakeImpl():
    '''Fake implementation of a model.'''
    dataset: Optional[Dataset] = None

    def __init__(self, **kwargs):
        pass

    def fit(self, dataset: Optional[Dataset]) -> None:
        self.dataset = dataset

    def predict(self, dataset: Optional[Dataset]) -> Optional[Dataset]:
        return dataset


class FakeFitter(Algorithm):
    '''Pretend to do something with data.'''
    _name = 'fake_model'

    def instantiate(self, **kwargs) -> 'FakeFitterInstance':
        return FakeFitterInstance(parent=self, **kwargs)


class FakeFitterInstance(SklearnLikeAlgorithmInstance):
    '''Pretend to fit data.'''
    _impl = FakeImpl()
    _constructor = FakeImpl

    def __init__(self, parent, **kwargs):
        super().__init__(parent=parent, **kwargs)

    def fit(self, dataset: Optional[Dataset]) -> None:
        '''Fit a model based on train data.'''
        self._impl.fit(dataset)
        self._trained = True

    @property
    def dataset(self) -> Optional[Dataset]:
        return self._impl.dataset


REFERENCE_DATASET = Dataset(
    a_key='a_value',
    another_key='another_value',
)


PREDICT_DATASET = Dataset(
    covariates='predict_value'
)


PREDICTED_DATASET = Dataset(
    predictions='predict_value'
)


def test_serialize_round_trip() -> None:
    dut = FakeFitter()
    inst = dut.instantiate()
    assert inst.dataset is None
    inst.fit(dataset=REFERENCE_DATASET)
    saved = inst.serialize()
    inst = dut.instantiate()
    loaded_inst = inst.deserialize(serialized_model=saved)
    assert isinstance(loaded_inst, FakeFitterInstance)
    assert loaded_inst.dataset == REFERENCE_DATASET


def test_predict_sunny_day() -> None:
    dut = FakeFitter()
    inst = dut.instantiate()

    inst.fit(dataset=REFERENCE_DATASET)
    got = inst.predict(dataset=PREDICT_DATASET)
    assert got == PREDICTED_DATASET



'''Tests for the Ranker class implementations'''
import re
from typing import Optional

import pandas as pd
import pytest

from ..algorithms.connect import ConnectorModel
from ..executor.executor import ExecutorKind
from ..generator.bound_pipeline import BoundPipelineStub, BoundPipeline
from ..generator.designator import Designator
from ..instantiator.executable_pipeline import (ExecutablePipelineStub,
                                                PipelineResult,
                                                PipelineResults)
from ..metrics.impl.metric import MetricInvalidDatasetError
from ..metrics.impl.metric_catalog import Metric
from ..wrangler.dataset import Dataset
from .ranker import RankerError, ScoredResult
from .ranker_impl import RankerImpl

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.


# pylint: disable=missing-function-docstring,duplicate-code
# pylint: disable=missing-class-docstring
# pylint: disable=too-few-public-methods
# pylint: disable=redefined-outer-name


class FakeMetricLow(Metric):
    '''Fake metric where lower is better'''
    _name = 'fake_metric_low'
    _tags = {'high': ['false']}

    def calculate(self,
                  pred: Dataset,
                  ground_truth: Optional[Dataset] = None) -> float:
        if 'column' not in pred:
            raise MetricInvalidDatasetError('hamster')
        return pred['column'][0]


class FakeMetricHigh(Metric):
    '''Fake metric where higher is better'''
    _name = 'fake_metric_high'
    _tags = {'high': ['true']}

    def calculate(self,
                  pred: Dataset,
                  ground_truth: Optional[Dataset] = None) -> float:
        if 'column' not in pred:
            raise MetricInvalidDatasetError('hamster')
        return pred['column'][0]


class FakeMetricAdd(Metric):
    '''Fake metric which adds .5 to whatever FakeMetricHigh returns'''
    _name = 'fake_metric_add'
    _tags = {'high': ['true']}

    def calculate(self,
                  pred: Dataset,
                  ground_truth: Optional[Dataset] = None) -> float:
        return FakeMetricHigh().calculate(pred, ground_truth) + 0.5


PIPELINE1 = ExecutablePipelineStub(
    kind=ExecutorKind('stub_executor_kind'),
    pipeline=BoundPipelineStub(name='pipeline1'))


PIPELINE2 = ExecutablePipelineStub(
    kind=ExecutorKind('stub_executor_kind'),
    pipeline=BoundPipelineStub(name='pipeline2'))


UNRANKABLE_PIPE = ExecutablePipelineStub(
    kind=ExecutorKind('stub_executor_kind'),
    pipeline=BoundPipelineStub(name='unrankable_pipe'))


ERROR_PIPE = ExecutablePipelineStub(
    kind=ExecutorKind('stub_executor_kind'),
    pipeline=BoundPipelineStub(name='error_pipe')
)


@pytest.fixture(scope='session')
def result1() -> PipelineResult:
    prediction1 = Dataset(column=[1.0])
    return PipelineResult(executable_pipeline=PIPELINE1,
                          prediction=prediction1)


@pytest.fixture(scope='session')
def results(result1: PipelineResult) -> PipelineResults:
    result2 = PipelineResult(executable_pipeline=PIPELINE2,
                             prediction=Dataset(column=[2.0]))
    unrankable_res = PipelineResult(executable_pipeline=UNRANKABLE_PIPE,
                                    prediction=Dataset())
    error_res = PipelineResult(executable_pipeline=ERROR_PIPE,
                               prediction=Dataset(
                                   {'error': 'gerbil'}))
    des1 = Designator('result1')
    des2 = Designator('result2')
    unrankable_des = Designator('unrankable_res')
    error_des = Designator('error')
    return PipelineResults({
        des1: result1,
        des2: result2,
        unrankable_des: unrankable_res,
        error_des: error_res})


def test_scored_result_lt_different_metrics(result1: PipelineResult) -> None:
    met1 = FakeMetricLow()
    met2 = FakeMetricHigh()
    res1 = ScoredResult(metric=met1,
                        result=result1)
    res2 = ScoredResult(metric=met2,
                        result=result1)
    with pytest.raises(RankerError,
                       match=r'(c_low.*c_high)|(c_high.*c_low)'):
        _ = res1 < res2


def test_unscorable_reason(result1: PipelineResult) -> None:
    res = ScoredResult(metric=FakeMetricHigh(),
                       result=result1,
                       unscorable_reason='foo')
    assert 'foo' in str(res)


def test_ranker_low(results: PipelineResults) -> None:
    dut = RankerImpl()
    got = dut.rank(results=results, metrics={'fake_metric_low': FakeMetricLow()})

    # low: expect [res1, res2, gerbil, hamster]
    assert re.search(
        pattern='pipeline1.*pipeline2.*error_pipe.*unrankable_pipe',
        string=str(got['fake_metric_low']),
        flags=re.MULTILINE | re.DOTALL)


def test_ranker_high(results: PipelineResults) -> None:
    dut = RankerImpl()
    got = dut.rank(results=results, metrics={'fake_metric_high': FakeMetricHigh()})

    # high: expect [res2, res1, gerbil, hamster]
    assert re.search(
        pattern='pipeline2.*pipeline1.*error_pipe.*unrankable_pipe',
        string=str(got),
        flags=re.MULTILINE | re.DOTALL)


def test_ranking_df(results: PipelineResults) -> None:
    dut = RankerImpl()
    got = dut.rank(results=results, metrics={'fake_metric_high': FakeMetricHigh()})
    want = pd.DataFrame({
        'pipeline': ['pipeline2', 'pipeline1', 'error_pipe', 'unrankable_pipe'],
        'fake_metric_high': [2.0, 1.0, 'gerbil', 'hamster']
    })
    pd.testing.assert_frame_equal(want, got['fake_metric_high'].as_dataframe())


def test_rankings_df(results: PipelineResults) -> None:
    dut = RankerImpl()
    got = dut.rank(results=results, metrics={
        'fake_metric_high': FakeMetricHigh(),
        'fake_metric_low': FakeMetricLow(),
        'fake_metric_add': FakeMetricAdd()
    })
    want = pd.DataFrame({
        'pipeline': ['pipeline2', 'pipeline1', 'error_pipe', 'unrankable_pipe'],
        'fake_metric_add': [2.5, 1.5, 'gerbil', 'hamster'],
        'fake_metric_high': [2.0, 1.0, 'gerbil', 'hamster'],
        'fake_metric_low': [2.0, 1.0, 'gerbil', 'hamster']
    })
    pd.testing.assert_frame_equal(want, got.as_dataframe())


def test_rankings_df_order(results: PipelineResults) -> None:
    dut = RankerImpl()
    got = dut.rank(results=results, metrics={
        'fake_metric_high': FakeMetricHigh(),
        'fake_metric_low': FakeMetricLow(),
        'fake_metric_add': FakeMetricAdd()
    })
    want = pd.DataFrame({
        'pipeline': ['pipeline1', 'pipeline2', 'error_pipe', 'unrankable_pipe'],
        'fake_metric_low': [1.0, 2.0, 'gerbil', 'hamster'],
        'fake_metric_add': [1.5, 2.5, 'gerbil', 'hamster'],
        'fake_metric_high': [1.0, 2.0, 'gerbil', 'hamster']
    })
    got_df = got.as_dataframe(order_metric='fake_metric_low')
    print(got_df)
    pd.testing.assert_frame_equal(want, got_df)


def test_family_score() -> None:
    # Create two pipelines in the same family, with different hyperparams
    pipeline1 = BoundPipeline(name='base_pipeline')
    pipeline1.step(model=ConnectorModel(name='model1', param1='some_value')).mark_queried()

    pipeline2 = BoundPipeline(name='base_pipeline')
    pipeline2.step(model=ConnectorModel(name='model1', param1='another_value')).mark_queried()

    # Create another pipeline with 2 queried connector models
    #   instead of one so it will not be in the same family
    other_family_pipe = BoundPipeline(name='base_pipeline')
    other_family_pipe.step(model=ConnectorModel(name='model1')).mark_queried()
    other_family_pipe.step(model=ConnectorModel(name='model2')).mark_queried()

    result1 = PipelineResult(
        executable_pipeline=ExecutablePipelineStub(pipeline=pipeline1),
        prediction=Dataset(column=[1.0]))

    result2 = PipelineResult(
        executable_pipeline=ExecutablePipelineStub(pipeline=pipeline2),
        prediction=Dataset(column=[2.0]))

    result_other_family = PipelineResult(
        executable_pipeline=ExecutablePipelineStub(pipeline=other_family_pipe),
        prediction=Dataset(column=[1.5]))

    results = PipelineResults({
        pipeline1.designator: result1,
        pipeline2.designator: result2,
        other_family_pipe.designator: result_other_family,
    })

    dut = RankerImpl()
    got = dut.rank(results=results, metrics={'fake_metric_high': FakeMetricHigh()})
    got_family_scores = got['fake_metric_high'].scores(all_scores=False)

    assert len(got_family_scores) == 2
    assert got_family_scores[0].pipeline_des == pipeline2.designator
    assert got_family_scores[1].pipeline_des == other_family_pipe.designator



'''Holds an instance of a fittable algorithm.'''
import abc
from copy import deepcopy
import pickle
from typing import Any, Optional
from typing_extensions import Protocol

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from ...wrangler.dataset import Dataset

from .algorithm_instance import AlgorithmInstance, Error


class UntrainedError(Error):
    '''Attempt to save the training of a model before it is trained.'''


class DeserializationError(Error):
    '''Attempt to deserialize a trained model failed'''


class FittableAlgorithmInstance(AlgorithmInstance, metaclass=abc.ABCMeta):
    '''Holds an instance of a fittable algorithm.'''
    _trained: bool = False

    @property
    def trained(self) -> bool:
        '''Is this algorithm trained?'''
        return self._trained

    def deserialize(self, serialized_model: bytes) -> 'FittableAlgorithmInstance':
        '''Restore a trained model.'''
        self._trained = True
        try:
            instance = pickle.loads(serialized_model)
        except pickle.UnpicklingError as err:
            raise DeserializationError(
                'Base AlgorithmInstance unexpectedly handed a serialized_model '
                'that is not a pickle. '
                'Is there a missing implementation of deserialize()?'
            ) from err
        if not isinstance(instance, FittableAlgorithmInstance):
            raise DeserializationError(
                f'Base AlgorithmInstance unexpectedly handed a {instance} '
                f'of type {type(instance)}, expected a subtype of AlgorithmInstance. '
                'Is there a missing implementation of deserialize()?')
        # The algorithm may contain catalogs which are not pickleable.
        instance._algorithm = self._algorithm   # pylint: disable=protected-access
        return instance

    def serialize(self) -> bytes:
        '''Return a serialized version of a trained model.

        By default, we save the whole AlgorithmInstance.'''
        if not self._trained:
            raise UntrainedError(f'attempt to serialize before fit for {self.catalog_name}')
        minime = deepcopy(self)
        # The algorithm may contain catalogs which are not pickleable.
        minime._algorithm = None  # pylint: disable=protected-access
        return pickle.dumps(minime)

    @abc.abstractmethod
    def fit(self, dataset: Optional[Dataset]) -> None:
        '''Fit a model based on train data.

        This should set self._trained to True
        '''

    @abc.abstractmethod
    def predict(self, dataset: Optional[Dataset]) -> Optional[Dataset]:
        '''Apply model to input dataset to create output.

        This may require that the model is fit (self.trained == True) before it is called.
        '''


class Constructor(Protocol):
    '''Match the signature of a generic model constructor.'''
    def __call__(self, **kwargs: Any) -> Any:
        ...


def not_implemented(**kwargs):
    '''A constructor must be specified in a subclass'''
    raise NotImplementedError('Implementation must specify a model _constructor')



'''Tests for single_fold_validator.py'''
# pylint: disable=missing-class-docstring,missing-function-docstring
# pylint: disable=super-init-not-called,unused-argument,duplicate-code

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from typing import Dict

from ..executor.executor import ExecutorStub
from ..generator.bound_pipeline import BoundPipeline
from ..generator.designator import Designator
from ..instantiator.executable_pipeline import (ExecutablePipeline,
                                                PipelineResult)
from ..instantiator.instantiator_factory import InstantiatorFactory
from ..splitters.impl.splitter import Fold, SplitDataset
from ..wrangler.dataset import Dataset
from .single_fold_validator import SingleFoldValidator


class FakeExecutor(ExecutorStub):

    def predict(self,
                dataset: Dataset,
                pipelines: Dict[Designator, ExecutablePipeline]
                ) -> Dict[Designator, PipelineResult]:
        '''Does not use pipeline, instead rewrites dataset to {'newkey': 'newval'}'''
        predictions = dataset.output()
        predictions['newkey'] = 'newval'
        return {d: PipelineResult(executable_pipeline=p, prediction=predictions)
                for d, p in pipelines.items()}


def test_sunny_day() -> None:
    pipe = BoundPipeline(name=Designator('des'), tags={})
    split_data = SplitDataset([Fold(
        train=Dataset({'k_tr': 'v_tr'}),
        validate=Dataset({'k_va': 'v_va'}))])
    dut = SingleFoldValidator()
    res = dut.validate_pipelines(
        split_dataset=split_data,
        bound_pipelines={Designator('des'): pipe},
        instantiator=InstantiatorFactory(),
        executor=FakeExecutor())

    assert Designator('des') in res
    single_res = res[Designator('des')]
    assert single_res.prediction == Dataset({'newkey': 'newval'})
    assert single_res.split_dataset is not None
    assert single_res.split_dataset.folds[0].train == Dataset({'k_tr': 'v_tr'})
    assert single_res.split_dataset.folds[0].validate == Dataset({'k_va': 'v_va'})



'''Tests for BoundPipeline'''

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

import pytest

from ..algorithms.impl.algorithm import AlgorithmStub
from ..generator.designator import Designator
from ..searcher.frozen_overrides import FrozenOverrides

from .bound_pipeline import BoundPipelineStub, ValidationError

# pylint: disable=missing-function-docstring,duplicate-code


def test_empty_pipeline_is_valid() -> None:
    dut = BoundPipelineStub(name='Fake_Template', tags={'DATATYPE': 'IMAGE'})
    dut.validate()


def test_step() -> None:
    dut = BoundPipelineStub(name='Fake_Template', tags={'DATATYPE': 'IMAGE'})
    dut.step(model=AlgorithmStub('A'))
    dut.step(model=AlgorithmStub('T'))
    dut.validate()


def test_invalid() -> None:
    dut = BoundPipelineStub(name='Fake_Template', tags={'DATATYPE': 'IMAGE'})
    dut.step()
    dut.step().set_name("step2")
    with pytest.raises(ValidationError, match='step2'):
        dut.validate()


def test_parallel() -> None:
    dut = BoundPipelineStub(name='Fake_Template')
    pipe1 = dut.new(name='at')
    pipe1.step(AlgorithmStub('A'))
    pipe1.step(AlgorithmStub('T'))
    pipe2 = dut.new(name='dt')
    pipe2.step(AlgorithmStub('D'))
    pipe2.step(AlgorithmStub('T'))
    dut.parallel(left=pipe1, right=pipe2)
    dut.validate()


def test_parallel_invalid() -> None:
    '''Confirm that errors on multiple pipelines are all reported.

    The errors here are both missing models, but on separate pipelines.
    '''
    dut = BoundPipelineStub(name='Fake_Template')
    pipe1 = dut.new(name='at')
    pipe1.step(AlgorithmStub('A'))
    pipe1.step().set_name('missing_t')
    pipe2 = dut.new(name='dt')
    pipe2.step().set_name('missing_d')
    pipe2.step(AlgorithmStub('T'))
    dut.parallel(left=pipe1, right=pipe2)
    with pytest.raises(ValidationError, match='missing_.*missing_'):
        dut.validate()


def test_designator_generation() -> None:
    dut = BoundPipelineStub(name='opt_name')
    dut.step(AlgorithmStub('query1')).mark_queried()
    dut.step(AlgorithmStub('nonquery2'))
    dut.step(AlgorithmStub('query3')).mark_queried()
    dut.step(AlgorithmStub('nonquery4'))

    assert str(dut.designator) == 'opt_name@query1@query3'

    frozen_overrides = FrozenOverrides.freeze({
        'query3': {'h': 'v'},
        'nonquery4': {'h1': 'v1', 'h2': 'v2'}})
    got = BoundPipelineStub.build(
        steps=dut.steps,
        template_name=Designator("new_template_name"),
        frozen_overrides=frozen_overrides)

    want = 'new_template_name@query1@query3:h=v@nonquery4:h1=v1,h2=v2'
    assert str(got.designator) == want

    want_family = 'new_template_name@query1@query3'
    assert str(got.family_designator) == want_family



# flake8: noqa
# type: ignore
# pylint: skip-file

# ngautonml's distributed AI implementation is adapted from this code made by Kyle Miller
# We are keeping it here for now to compare and make sure we adapted it right.
# There is a test that compares the ngautonml implementation.

from typing import Any

class DistAIModel:
    def __init__(self):
        self.fit_time = None
        self.ID = None

    def serialize(self): # returns serialized model
        # care should be taken to make message size as small as possible. self.ID is required.
        # the deserialized model must be able to run norm2, inner_product, and metrics. It need not run fit.
        pass

    @staticmethod
    def deserialize(serialized_model): # returns model instance
        # care should be taken to make message size as small as possible. self.ID is required.
        # the deserialized model must be able to run norm2, inner_product, and metrics. It need not run fit.
        pass

    def fit(self, data, neighbor_models): # returns nothing, fits model to data and other models
        pass

    def norm2(self): # returns scalar
        pass

    def inner_product(self,other): # returns scalar
        pass

    def metrics(self,data): # returns a dictionary with string keys recording any sort of metrics Dict[str, Any]
        pass

def register(catalog: Any, *args, **kwargs) -> None:
    '''Register all the objects in this file.'''



'''Tests for the "simple" executor'''
import re
from typing import Dict, Optional, Tuple

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

import logging
import pytest
import numpy as np
import pandas as pd
from sklearn import datasets  # type: ignore[import]

from ...algorithms.impl.algorithm import Algorithm
from ...algorithms.impl.fittable_algorithm_instance import FittableAlgorithmInstance
from ...algorithms.impl.algorithm_auto import AlgorithmCatalogAuto
from ...generator.bound_pipeline import BoundPipeline
from ...generator.designator import Designator, StepDesignator
from ...instantiator.executable_pipeline import (ExecutablePipeline,
                                                 FitError, PredictError)
from ...algorithms.connect import ConnectorModel
from ...wrangler.dataset import Dataset, Metadata, RoleName
from ...wrangler.dataset import DatasetKeys, Column

from ..cucumber import JarOfCucumbers

from .simple_instantiator import SimpleInstantiator
from .simple_executor import SimpleExecutor
from .simple_executable_pipeline import SimpleExecutablePipeline

# pylint: disable=missing-function-docstring,missing-class-docstring,duplicate-code,protected-access

REFERENCE_DATASET = Dataset(
    a_key='a_value',
    another_key='another_value',
)


def make_pipeline(instantiator: SimpleInstantiator, name: Designator, **kwargs):
    pipe = BoundPipeline(name=name, tags={})
    connector = ConnectorModel()
    pipe.step(model=connector, new_key='a_key', another_key='another_key').set_name('connector1.1')
    pipe.step(model=connector, **kwargs)

    return instantiator.instantiate(pipeline=pipe)


def load_datasets() -> Tuple[Dataset, Dataset, Dataset]:
    # Load the diabetes dataset
    diabetes_x, diabetes_y = datasets.load_diabetes(return_X_y=True)

    # Use only one feature
    diabetes_x = diabetes_x[:, np.newaxis, 2]

    # Split the data into training/testing sets
    diabetes_x_train = diabetes_x[:-20]
    diabetes_x_test = diabetes_x[-20:]

    # Split the targets into training/testing sets
    diabetes_y_train = diabetes_y[:-20]
    diabetes_y_test = diabetes_y[-20:]

    dataset_train = Dataset(
        metadata=Metadata(
            roles={RoleName.TARGET: [Column('target')]}
        ),
        covariates=pd.DataFrame(diabetes_x_train, columns=['attribute']),
        target=pd.DataFrame({'target': diabetes_y_train})
    )
    dataset_test = Dataset(
        metadata=Metadata(
            roles={RoleName.TARGET: [Column('target')]}
        ),
        covariates=pd.DataFrame(diabetes_x_test, columns=['attribute'])
    )
    ground_truth = Dataset(
        metadata=Metadata(
            roles={RoleName.TARGET: [Column('target')]}
        ),
        ground_truth=pd.DataFrame({'target': diabetes_y_test})
    )

    return (dataset_train, dataset_test, ground_truth)


def make_regression_pipeline(instantiator: SimpleInstantiator,
                             name: Designator
                             ) -> SimpleExecutablePipeline:
    catalog = AlgorithmCatalogAuto()
    regression = catalog.lookup_by_name('sklearn.linear_model.LinearRegression')
    connector = ConnectorModel()
    pipe = BoundPipeline(name=name, tags={})
    pipe.step(model=regression).set_name('regression')
    pipe.step(model=connector, result=DatasetKeys.PREDICTIONS.value)
    return instantiator.instantiate(pipeline=pipe)


def strip_serial(key: StepDesignator) -> str:
    return re.sub('[0-9]+', 'NUM', key)


def test_fit_sunny_day() -> None:
    instantiator = SimpleInstantiator()
    des1 = Designator('foo')
    des2 = Designator('bar')
    executable1 = make_regression_pipeline(
        instantiator=instantiator,
        name=des1)
    pipe = BoundPipeline(name=des2, tags={})
    pipe.step(model=ConnectorModel(),
              new_key=DatasetKeys.COVARIATES.value,
              another_key=DatasetKeys.TARGET.value).set_name('connector_special_name')
    executable2 = instantiator.instantiate(pipe)
    pipelines: Dict[Designator, ExecutablePipeline] = {des1: executable1, des2: executable2}

    train, _, _ = load_datasets()

    dut = SimpleExecutor()

    got_training = dut.fit(dataset=train, pipelines=pipelines)
    jar = got_training.models(name=des2)
    assert isinstance(jar, JarOfCucumbers)
    assert set({strip_serial(key) for key in jar.keys()}
               ) == set({'@connect_NUM@connector_special_name'})
    # TODO(Merritt/Piggy): assert more here once this change is more complete


def test_predict_sunny_day() -> None:
    instantiator = SimpleInstantiator()
    des1 = Designator('foo')
    des2 = Designator('bar')
    executable1 = make_pipeline(instantiator=instantiator,
                                name=des1,
                                twice_transformed_key='new_key',
                                another_key='another_key')
    executable2 = make_pipeline(instantiator=instantiator,
                                name=des2,
                                twice_transformed_key2='new_key')

    pipelines: Dict[Designator, ExecutablePipeline] = {des1: executable1, des2: executable2}

    dut = SimpleExecutor()

    dut.fit(dataset=REFERENCE_DATASET, pipelines=pipelines)

    result = dut.predict(dataset=REFERENCE_DATASET, pipelines=pipelines)

    assert [r.prediction for r in result.values()] == [
        {'twice_transformed_key': 'a_value', 'another_key': 'another_value'},
        {'twice_transformed_key2': 'a_value'},
    ]


def test_parallel_sunny_day() -> None:
    instantiator = SimpleInstantiator()
    pipe = BoundPipeline(name=Designator('parallel_pipeline'), tags={})
    subpipe1 = pipe.new('connector_pipeline')
    subpipe1.step(model=ConnectorModel(), new_key='covariates').set_name('connector_step')
    subpipe2 = pipe.new('regression_pipeline')
    catalog = AlgorithmCatalogAuto()
    model = catalog.lookup_by_name('sklearn.linear_model.LinearRegression')
    subpipe2.step(model=model).set_name('regression_step')
    subpipe2.step(model=ConnectorModel(), result='predictions'
                  ).set_name('regression_connect_step')
    pipe.parallel(connector=subpipe1, regression=subpipe2).set_name('parallel_step')
    executable = instantiator.instantiate(pipe)

    train, test, _ = load_datasets()

    dut = SimpleExecutor()

    fake_des = Designator('fake_des')

    dut.fit(dataset=train,
            pipelines={fake_des: executable})

    got_predict = dut.predict(dataset=test,
                              pipelines={fake_des: executable})

    got_result = list(got_predict.values())[0]
    assert got_result.prediction is not None
    got_connector = got_result.prediction['connector']
    got_regression = got_result.prediction['regression']

    assert set(got_connector.keys()) == set(['new_key'])
    assert got_connector['new_key'].iat[0, 0] == pytest.approx(0.07786338762689, 1e-6)
    assert set(got_regression.keys()) == set(['result'])
    assert got_regression['result'].iat[0, 0] == pytest.approx(225.9732401, 1e-6)


class BrokenAlg(Algorithm):

    def instantiate(self, **hyperparams) -> FittableAlgorithmInstance:
        if hyperparams['kind'] == 'fit':
            del hyperparams['kind']
            return BrokenFitAlgInstance(parent=self, **hyperparams)
        if hyperparams['kind'] == 'predict':
            del hyperparams['kind']
            return BrokenPredictAlgInstance(parent=self, **hyperparams)
        raise NotImplementedError(f'Unimplemented kind: {hyperparams["kind"]}')


class BrokenFitAlgInstance(FittableAlgorithmInstance):
    _name = 'fit is broken'

    def fit(self, dataset: Optional[Dataset]) -> None:
        _ = 1 / 0

    def predict(self, dataset: Optional[Dataset]) -> Optional[Dataset]:
        return dataset


class BrokenPredictAlgInstance(FittableAlgorithmInstance):

    def fit(self, dataset: Optional[Dataset]) -> None:
        self._trained = True

    def predict(self, dataset: Optional[Dataset]) -> Optional[Dataset]:
        # we need this condition because the model's predict() gets called as part
        #   of the pipeline's fit() call, however we only want to raise an error
        #   on the pipeline's predict() call
        if 'hamster' in (dataset or {}):
            raise NotImplementedError('hamster')
        return dataset


def test_fit_error(caplog: pytest.LogCaptureFixture) -> None:
    pipe = BoundPipeline(name=Designator('broken_fit'))
    pipe.step(model=BrokenAlg(name='fit is broken', kind='fit'))
    executable = SimpleInstantiator().instantiate(pipeline=pipe)

    dut = SimpleExecutor()
    with caplog.at_level(logging.WARNING):
        got = dut.fit(
            dataset=Dataset(),
            pipelines={Designator('broken_fit'): executable})

    got_pipe = got[Designator('broken_fit')]
    assert isinstance(got_pipe, FitError)
    assert 'simple_executor_test.py' in str(got_pipe)
    assert 'ZeroDivisionError' in str(got_pipe)
    assert 'ZeroDivisionError' in caplog.text
    assert 'broken_fit' in caplog.text


def test_predict_error(caplog: pytest.LogCaptureFixture) -> None:
    pipe = BoundPipeline(name=Designator('broken_predict'))
    pipe.step(model=BrokenAlg(name='predict is broken', kind='predict'))
    executable = SimpleInstantiator().instantiate(pipeline=pipe)

    dut = SimpleExecutor()

    des = Designator('broken_predict')
    pipes: Dict[Designator, ExecutablePipeline] = {des: executable}
    dut.fit(dataset=Dataset(), pipelines=pipes)
    with caplog.at_level(logging.WARNING):
        got = dut.predict(
            dataset=Dataset({'hamster': 'something'}),
            pipelines=pipes)
    got_pipe = got[des].prediction
    assert got_pipe is not None
    assert isinstance(got_pipe['error'], PredictError)
    assert 'simple_executor_test.py' in str(got_pipe['error'])
    assert 'NotImplementedError' in str(got_pipe['error'])
    assert 'hamster' in str(got_pipe['error'])
    assert 'hamster' in caplog.text
    assert 'broken_predict' in caplog.text


def test_predict_after_fit_error() -> None:
    '''If we get a fit error, we want pipe.predict() to output that fit error
    inside its prediction dataset.
    '''
    pipe = BoundPipeline(name=Designator('broken_fit'))
    pipe.step(model=BrokenAlg(name='fit is broken', kind='fit'))
    executable = SimpleInstantiator().instantiate(pipeline=pipe)

    des = Designator('broken_fit')
    pipes: Dict[Designator, ExecutablePipeline] = {des: executable}

    dut = SimpleExecutor()
    dut.fit(dataset=Dataset(),
            pipelines=pipes)

    got = dut.predict(dataset=Dataset(),
                      pipelines=pipes)

    got_pipe = got[des].prediction
    assert got_pipe is not None
    assert isinstance(got_pipe['error'], FitError)
    assert 'simple_executor_test.py' in str(got_pipe['error'])
    assert 'ZeroDivisionError' in str(got_pipe['error'])



'''Tests ensuring that the problem definition JSON in the examples folder is valid.'''
import os
from pathlib import Path

from ..problem_def.problem_def import ProblemDefinition

# pylint: disable=missing-function-docstring,duplicate-code


def module_path() -> Path:
    current_path = str(os.getenv('PYTEST_CURRENT_TEST')).split('::', maxsplit=1)[0]
    pathobj = Path(current_path).resolve()
    module_parent = pathobj.parents[2]
    return module_parent


def test_example_tabular_regression(tmp_path: Path) -> None:
    path = module_path() / 'examples' / 'regression' / 'diabetes.json'
    csv_path = module_path() / 'examples' / 'regression' / 'diabetes-train.csv'
    test_csv_path = module_path() / 'examples' / 'regression' / 'diabetes-test.csv'
    with path.open() as file:
        pd_str = file.read()
    pd_paths_fixed = pd_str.replace(
        'ngautonml/examples/regression/diabetes-train.csv',
        str(csv_path)
    )
    pd_paths_fixed = pd_paths_fixed.replace(
        'ngautonml/examples/regression/diabetes-test.csv',
        str(test_csv_path)
    )
    pd_paths_fixed = pd_paths_fixed.replace(
        'ngautonml/examples/regression/diabetes-output',
        str(tmp_path / 'output_dir')
    )
    ProblemDefinition(pd_paths_fixed)  # fails if example json is not valid


def test_example_tabular_classification(tmp_path: Path):
    pd_path = module_path() / 'examples' / 'classification' / 'credit.json'
    csv_path = module_path() / 'examples' / 'classification' / 'credit-train.csv'
    test_csv_path = module_path() / 'examples' / 'classification' / 'credit-test.csv'
    with pd_path.open() as file:
        pd_str = file.read()
    pd_paths_fixed = pd_str.replace(
        'ngautonml/examples/classification/credit-train.csv',
        str(csv_path)
    )
    pd_paths_fixed = pd_paths_fixed.replace(
        'ngautonml/examples/classification/credit-test.csv',
        str(test_csv_path)
    )
    pd_paths_fixed = pd_paths_fixed.replace(
        'ngautonml/examples/classification/credit-output',
        str(tmp_path / 'output_dir')
    )
    ProblemDefinition(pd_paths_fixed)  # fails if example json is not valid


def test_example_tabular_classification_arff(tmp_path: Path):
    pd_path = module_path() / 'examples' / 'classification' / 'credit-arff.json'
    arff_path = module_path() / 'examples' / 'classification' / 'dataset_31_credit-g.arff'
    with pd_path.open() as file:
        pd_str = file.read()
    pd_paths_fixed = pd_str.replace(
        'ngautonml/examples/classification/dataset_31_credit-g.arff',
        str(arff_path)
    )
    pd_paths_fixed = pd_paths_fixed.replace(
        'ngautonml/examples/classification/credit-output',
        str(tmp_path / 'output_dir')
    )
    ProblemDefinition(pd_paths_fixed)  # fails if example json is not valid



'''Decider that measures elapsed time since last send to a specific neighbor.'''
# Copyright (c) 2024 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

import time
from typing import Mapping

from ..algorithms.impl.neighbor_metadata_interface import NeighborMetadataInterface
from ..config_components.distributed_config import DecidersConfig
from ..neighbor_manager.node_id import NodeID  # type: ignore[import-untyped]

from .impl.decider import Decider
from .impl.decider_catalog import DeciderCatalog

from .time_since_sent import TimeSinceSentDeciderConfig


class TimeSinceSentToDeciderConfig(TimeSinceSentDeciderConfig):
    '''Configuration for TimeSinceSentToDecider.'''


class TimeSinceSentToDecider(Decider):
    '''Decider that triggers a refractory time past the last sent message.'''
    name = 'time_since_sent_to'
    tags = {}
    _config: TimeSinceSentToDeciderConfig

    def decide(self, my_meta: NeighborMetadataInterface,
               neighbor_id: NodeID, neighbors: Mapping[NodeID, NeighborMetadataInterface]) -> bool:
        '''Decide if we should send to neighbor.'''
        assert my_meta.current_state is not None
        interval = time.monotonic() - neighbors[neighbor_id].last_state_sent_time
        return interval > self._config.time_seconds


def register(catalog: DeciderCatalog) -> None:
    '''Register all objects in this file.'''
    catalog.register(TimeSinceSentToDecider)
    DecidersConfig.register(name=TimeSinceSentToDecider.name,
                            config_type=TimeSinceSentToDeciderConfig)



'''Load data from a given format to a given format.'''

# Copyright (c) 2024 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

import abc
import inspect
from typing import Any, Dict, List, Optional

from ...config_components.dataset_config import DatasetConfig
from ...wrangler.dataset import Dataset, Metadata


class DataLoader(metaclass=abc.ABCMeta):
    '''Base class for things that load data from a given format to a given format.'''
    name: str = 'data_loader_base_class'
    tags: Dict[str, List[str]] = {
        'input_format': [],
        'loaded_format': [],
    }
    _config: DatasetConfig
    _metadata: Metadata

    def __init__(self, config: DatasetConfig):
        self._config = config
        self._metadata = config.metadata

    @abc.abstractmethod
    def validate(self, dataset: Optional[Dataset]) -> None:
        '''Raise a ValidationErrors if something about the dataset is inconsistent.'''

    @abc.abstractmethod
    def _load_train(self) -> Optional[Dataset]:
        ...

    def load_train(self) -> Optional[Dataset]:
        '''Load a Dataset from the information in the problem description'''
        retval = self._load_train()
        self.validate(retval)
        return retval

    @abc.abstractmethod
    def _load_test(self) -> Optional[Dataset]:
        ...

    def load_test(self) -> Optional[Dataset]:
        '''Load test data if it exists, otherwise None'''
        retval = self._load_test()
        if retval is not None:
            self.validate(retval)
        return retval

    @abc.abstractmethod
    def _load_ground_truth(self) -> Optional[Dataset]:
        '''If the target column exists in the test set, extract it as the ground truth.'''

    def load_ground_truth(self) -> Optional[Dataset]:
        '''If the target column exists in the test set, extract it as the ground truth.'''
        retval = self._load_ground_truth()
        if retval is not None:
            self.validate(retval)
        return retval

    @abc.abstractmethod
    def _dataset(self, data: Any, **kwargs) -> Dataset:
        ...

    @abc.abstractmethod
    def build_dataset_from_json(self, json_data: Dict) -> Dataset:
        '''Build a Dataset object from a JSON object.'''

    def dataset(self, data: Any, **kwargs) -> Dataset:
        '''Load a Dataset object with the same metadata, by placing data at supplied key.'''
        retval = self._dataset(data=data, **kwargs)
        self.validate(retval)
        return retval

    @abc.abstractmethod
    def poll(self, timeout: Optional[float] = 0.0) -> Optional[Dataset]:
        '''Poll for new training data.

        We'll wait up to timeout seconds for new data to arrive.
        If it times out we'll return None.

        A timeout of None waits forever.

        In a distributed simulation, the data is split based on node ID.
        '''

    def _lookup_var(self, var, default=None) -> Any:
        '''Look up a variable in the call stack.'''
        for frameinfo in inspect.stack(0)[:]:
            if var in frameinfo.frame.f_locals:
                return frameinfo.frame.f_locals[var]
        return default



'''Tests for params.py'''

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

import pytest

from ..wrangler.constants import RangeMethod

from .params import Matcher
from .params import ParamRange, ParamRanges, Selector, Override, Overrides

# pylint: disable=missing-function-docstring,duplicate-code


SAMPLE_OVERRIDE = Override(
    selector=Selector({
        Matcher.ALGORITHM: 'foo_model'
    }),
    params=ParamRanges(
        some_param=ParamRange(
            method=RangeMethod.FIXED,
            prange='some_param_value'
        )
    )
)


def test_overrides():
    dut = Overrides()
    dut.append(SAMPLE_OVERRIDE)

    assert len(dut) == 1
    for override in dut:
        assert override.selector[Matcher('algorithm')] == 'foo_model'
        assert override.params['some_param'].range == 'some_param_value'


def test_override_failures():
    dut = Overrides()

    with pytest.raises(TypeError, match='not subscriptable'):
        dut[0]  # pylint: disable=pointless-statement, unsubscriptable-object

    with pytest.raises(TypeError, match='does not support item assignment'):
        dut[0] = SAMPLE_OVERRIDE  # pylint: disable=unsupported-assignment-operation



'''Tests for auton_mean with deciders'''

# Copyright (c) 2024 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

# pylint: disable=missing-function-docstring, duplicate-code, protected-access

import pandas as pd
import pytest


from ..algorithms.impl.synchronous import advance
from ..config_components.distributed_config import DistributedConfig
from ..conftest import Clearer
from ..deciders.impl.decider_auto import DeciderCatalogAuto
from ..wrangler.base_port import BasePort
from ..wrangler.dataset import Dataset
from ..wrangler.logger import Logger

from ..algorithms.distributed.auton_mean import AutonMean

_ = DeciderCatalogAuto()  # pylint: disable=pointless-statement

logger = Logger(__file__).logger()

base_port = BasePort()


receiver_train = Dataset(
    covariates=pd.DataFrame({'a': [1, 2, 3], 'b': [2, 3, 4], 'c': [3, 4, 5]}))

sender_train = Dataset(
    covariates=pd.DataFrame({'a': [7, 8, 9], 'b': [8, 9, 10], 'c': [9, 10, 11]}))

empty_dataset = receiver_train.output()


LOOPBACK = '127.0.0.1'

TEST_TELEPHONE_N = 5
TEST_TELEPHONE_PORTS = [base_port.next() for _ in range(TEST_TELEPHONE_N)]


@pytest.mark.parametrize('strategy,distance_enabled', [
    ('broadcast', False),
    ('broadcast', True),
    ('unicast', False),
    ('unicast', True),
])
def test_telephone(strategy, distance_enabled,
                   assert_no_exceptions: Clearer) -> None:
    '''Whisper down the lane test. Nodes connected in a line.
    The only node that has data is one at the end '''
    n = TEST_TELEPHONE_N

    adjacency = {f'{k}': [k - 1, k + 1] for k in range(2, n)}
    adjacency['1'] = [2]
    adjacency[f'{n}'] = [n - 1]

    distributed_clause = {
        'discoverer': {
            'name': 'static',
            'static': {
                'adjacency': adjacency
            }
        },
        'communicator': {
            'name': 'memory',
            'memory': {
                'domain': f'test_telephone_{strategy}_{distance_enabled}',
            },
            'strategy': strategy,
        },
        'decider': {
            'distance': {
                'threshold': 0.1,
                'enabled': distance_enabled,
            },
        },
        'my_id': 0,
    }

    distributed_clauses = [dict(distributed_clause, my_id=k) for k in range(1, n + 1)]

    alg = AutonMean()

    telephone_params = dict(alg._default_hyperparams, Lambda=1000000)

    params = [dict(telephone_params) if k == 0 else dict(telephone_params) for k in range(n)]

    duts = [alg.instantiate(
        distributed=DistributedConfig(dc),
        synchronous=True,
        **ps) for (dc, ps) in zip(distributed_clauses, params)]
    try:
        for dut in duts:
            dut.start()

        duts[0].fit(receiver_train)

        # Wait for the message to arrive and get processed.
        for _ in range(3 * TEST_TELEPHONE_N):
            advance(duts)

        # Confirm that neighbor optimization doesn't mess up result.
        got_predictions = [dut.predict(empty_dataset) for dut in duts]

        for got_prediction in got_predictions:
            assert got_prediction is not None
            pd.testing.assert_frame_equal(
                got_prediction.predictions,
                pd.DataFrame({'a': [2.0], 'b': [3.0], 'c': [4.0]}))

        assert_no_exceptions(duts)

    finally:
        for dut in duts:
            dut.stop()


TEST_N_SAME_N = 5
TEST_N_SAME_PORTS = [base_port.next() for _ in range(TEST_N_SAME_N)]


@pytest.mark.parametrize('strategy,distance_enabled', [
    ('broadcast', False),
    ('broadcast', True),
    ('unicast', False),
    ('unicast', True),
])
def test_n_same(strategy, distance_enabled,
                assert_no_exceptions: Clearer) -> None:
    '''N nodes fully connected with the same training match the non-distributed result.'''
    n = TEST_N_SAME_N

    adjacency = {f'{k}': [j for j in range(1, n + 1) if j != k] for k in range(1, n + 1)}

    distributed_clause = {
        'discoverer': {
            'name': 'static',
            'static': {
                'adjacency': adjacency
            }
        },
        'communicator': {
            'name': 'memory',
            'memory': {
                'domain': f'test_n_same_{strategy}_{distance_enabled}',
            },
            'strategy': strategy,
        },
        'decider': {
            'distance': {
                'threshold': 0.1,
                'enabled': distance_enabled,
            },
        },
        'my_id': 0,
    }

    distributed_clauses = [dict(distributed_clause, my_id=k) for k in range(1, n + 1)]

    alg = AutonMean()

    duts = [alg.instantiate(
        distributed=DistributedConfig(dc), synchronous=True) for dc in distributed_clauses]

    try:
        for dut in duts:
            dut.start()

        for dut in duts:
            dut.fit(receiver_train)

        # Wait for the messages to arrive and get processed.
        for _ in range(3):
            advance(duts)

        # Confirm that neighbor optimization doesn't mess up result.
        got_predictions = [dut.predict(empty_dataset) for dut in duts]

        for got_prediction in got_predictions:
            assert got_prediction is not None
            pd.testing.assert_frame_equal(
                got_prediction.predictions,
                pd.DataFrame({'a': [2.0], 'b': [3.0], 'c': [4.0]}))

        assert_no_exceptions(duts)

    finally:
        for dut in duts:
            dut.stop()



'''Autoloader for the splitter catalog.'''

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from pathlib import Path

from ...catalog.plugin_catalog import PluginCatalog

from .splitter import SplitterCatalog, Splitter


class SplitterCatalogAuto(PluginCatalog[Splitter], SplitterCatalog):
    '''Splitter catalog that automatically loads all splitters from a directory.

    Data splitters are in ../splitters.
    '''

    def __init__(self, **kwargs):
        super().__init__(catalog_name='splitters', default_root=Path(__file__).parents[2], **kwargs)



'''Tests for dataframe.py.'''
import numpy as np
import pandas as pd

from .impl.table_auto import TableCatalogAuto
from .impl.table import TableFactory

from .dataframe import DataFrameTable


def test_dataframe_table():
    '''Test the DataFrameTable.'''
    _ = TableCatalogAuto()

    my_df = pd.DataFrame({'a': [1, 2, 3], 'b': [4, 5, 6]})
    my_table = TableFactory(my_df)  # pylint: disable=abstract-class-instantiated

    # assert isinstance(my_table, Table)
    assert isinstance(my_table, DataFrameTable)

    assert my_table['a'].tolist() == [1, 2, 3]
    assert my_table['b'].tolist() == [4, 5, 6]

    my_table['c'] = my_table['a'] + my_table['b']
    assert my_table['c'].tolist() == [5, 7, 9]

    assert my_table.as_(pd.DataFrame).iloc[0, 2] == 5

    my_ndarray = my_table.as_(np.ndarray)
    assert isinstance(my_ndarray, np.ndarray)
    assert my_ndarray.shape == (3, 3)



'''Generate port numbers for tests.'''

# Copyright (c) 2024 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

import os
import threading

import filelock

from .logger import Logger, Level

log = Logger(from_filename=__file__, to_filename=__file__, level=Level.VERBOSE).logger()

START_BASE_PORT = 49200
MAX_PORT = 65535
LOCK_FILE = '/var/lock/base_port.lock'
PORT_NUMBER = '/var/lock/base_port_port'


class BasePort():
    '''Generate unique port numbers for a test file.'''
    _base_port: int = START_BASE_PORT  # class property
    _base_port_inc: int = 100   # class property
    _lock_base_port: filelock.FileLock = filelock.FileLock(
        LOCK_FILE, thread_local=False)  # class property
    _lock_next_port: threading.Lock  # instance property
    _current_port: int  # instance property

    def __init__(self) -> None:
        with self._lock_base_port:
            self._current_port = self._next_base_port()
        self._lock_next_port = threading.Lock()

    @classmethod
    def _next_base_port(cls) -> int:
        if not os.path.exists(PORT_NUMBER):
            log.log(Level.VERBOSE, 'Creating port file at %s', PORT_NUMBER)
            with open(PORT_NUMBER, 'w', encoding='utf-8') as port_file:
                port_file.write(f'{START_BASE_PORT}\n')

        with open(PORT_NUMBER, 'r', encoding='utf-8') as port_file:
            line = port_file.readline()
            cls._base_port = int(line)
            if cls._base_port + cls._base_port_inc > MAX_PORT:
                log.log(Level.VERBOSE, "Resetting baseport")
                cls._base_port = START_BASE_PORT

        cls._base_port += cls._base_port_inc

        with open(PORT_NUMBER, 'w', encoding='utf-8') as port_file:
            port_file.write(f'{cls._base_port}\n')

        log.log(Level.VERBOSE, 'Creating base port on %s', cls._base_port)

        return cls._base_port

    def next(self) -> int:
        '''Return the next port number.'''
        with self._lock_next_port:
            retval = self._current_port
            log.log(Level.VERBOSE, 'next port: %s', retval)
            self._current_port += 1
        return retval



'''Tests for memory_communicator.py.'''

# Copyright (c) 2024 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

# pylint: disable=missing-function-docstring
# pylint: disable=duplicate-code, redefined-outer-name


from queue import Queue
from typing import Dict, Set

from ..config_components.distributed_config import DistributedConfig
from ..neighbor_manager.event import NewNeighbor, Recv
from ..neighbor_manager.node_id import NodeID

from .impl.communicator_auto import CommunicatorCatalogAuto
from .memory_communicator import MemoryCommunicator

_ = CommunicatorCatalogAuto()


def test_exchange() -> None:
    '''2 MemoryCommunicator on different nodes talk and listen to each other.'''

    known_neighbors0: Set[NodeID] = set()
    known_neighbors1: Set[NodeID] = set()

    distributed = DistributedConfig(clause={
        'communicator': {
            'name': 'memory',
            'memory': {
                'domain': 'test_exchange',
            },
        }
    })

    dut0 = MemoryCommunicator(
        my_id=NodeID(0),
        known_neighbors=known_neighbors0,
        distributed=distributed
    )
    dut1 = MemoryCommunicator(
        my_id=NodeID(1),
        known_neighbors=known_neighbors1,
        distributed=distributed
    )

    queue0: Queue = Queue()
    queue1: Queue = Queue()

    dut0.start(queue=queue0)
    dut1.start(queue=queue1)

    try:
        dut0.send(dest_ids=[NodeID(1)], payload=b'Hello, Node 1!')
        first_event = queue1.get(block=True, timeout=0.5)
        assert isinstance(first_event, NewNeighbor)
        assert first_event.neighbor == NodeID(0)
        assert known_neighbors1 == {NodeID(0)}

        second_event = queue1.get(block=True, timeout=0.5)
        assert isinstance(second_event, Recv)
        assert second_event.neighbor == NodeID(0)
        assert second_event.payload == b'Hello, Node 1!'

        dut1.send(dest_ids=[NodeID(0)], payload=b'Nice to meet you, Node 0 :)')

        first_event = queue0.get(block=True, timeout=0.5)
        assert isinstance(first_event, NewNeighbor)
        assert first_event.neighbor == NodeID(1)
        assert known_neighbors0 == {NodeID(1)}

        second_event = queue0.get(block=True, timeout=0.5)
        assert isinstance(second_event, Recv)
        assert second_event.neighbor == NodeID(1)
        assert second_event.payload == b'Nice to meet you, Node 0 :)'

    finally:
        dut0.stop()
        dut1.stop()


def test_sendall() -> None:
    '''One node sends to a bunch of neighbors.'''

    known_neighbors: Dict[NodeID, Set[NodeID]] = {}
    distributed = DistributedConfig(clause={
        'communicator': {
            'name': 'memory',
            'memory': {
                'domain': 'test_sendall',
            },
        }
    })

    dut: Dict[NodeID, MemoryCommunicator] = {}
    queue: Dict[NodeID, Queue] = {}
    for node_id in range(0, 5):
        node = NodeID(node_id)
        known_neighbors[node] = set()
        queue[node] = Queue()
        dut[node] = MemoryCommunicator(
            my_id=node,
            known_neighbors=known_neighbors[node],
            distributed=distributed)
        dut[node].start(queue=queue[node], timeout=0.5)

    for node_id in range(1, 5):
        node = NodeID(node_id)
        # We need to send an initial message to establish our neighborliness.
        dut[node].send([NodeID(0)], f'Hello, I am {node}.'.encode())

    try:
        # Read a NewNeighbor and a Recv for each neighbor.
        # The pairs of requests are randomly interleaved due to threading,
        # so we don't bother checking what they are.
        for _ in range(1, 5):
            _ = queue[NodeID(0)].get()
            _ = queue[NodeID(0)].get()

        # This is what we are testing.
        dut[NodeID(0)].send_all(payload=b'Hello, All Nodes!')

        # Confirm that the test worked.
        for node_id in range(1, 5):
            node = NodeID(node_id)

            first_event = queue[node].get(block=True, timeout=0.5)
            assert isinstance(first_event, NewNeighbor)
            assert first_event.neighbor == NodeID(0)
            assert known_neighbors[node] == {NodeID(0)}

            second_event = queue[node].get(block=True, timeout=0.5)
            assert isinstance(second_event, Recv)
            assert second_event.neighbor == NodeID(0)
            assert known_neighbors[node] == {NodeID(0)}
            assert second_event.payload == b'Hello, All Nodes!'

    finally:
        for node_id in range(0, 5):
            node = NodeID(node_id)
            dut[node].stop()


def test_dynamic_discovery() -> None:
    '''One node sends to a bunch of neighbors.'''

    known_neighbors: Dict[NodeID, Set[NodeID]] = {}
    distributed = DistributedConfig(clause={
        'communicator': {
            'name': 'memory',
            'memory': {
                'domain': 'test_dynamic_discovery',
            },
        }
    })

    dut: Dict[NodeID, MemoryCommunicator] = {}
    queue: Dict[NodeID, Queue] = {}
    for node_id in range(0, 5):
        node = NodeID(node_id)
        known_neighbors[node] = set()
        queue[node] = Queue()
        dut[node] = MemoryCommunicator(
            my_id=node,
            known_neighbors=known_neighbors[node],
            distributed=distributed)
        dut[node].start(queue=queue[node], timeout=0.5)

    # This is the test. Everybody introduces themselves.
    for node_id in range(1, 5):
        node = NodeID(node_id)
        dut[node].send([NodeID(0)], f'Hello, I am {node}.'.encode())

    # Confirm that the test worked.
    try:
        # Read a NewNeighbor and a Recv for each neighbor.
        # The pairs of requests are randomly interleaved due to threading,
        # so we don't bother checking what they are.
        # This is needed to make sure every message has been recieved before
        #   checking known_neighbors.
        for _ in range(1, 5):
            _ = queue[NodeID(0)].get()
            _ = queue[NodeID(0)].get()

        assert known_neighbors[NodeID(0)] == {
            NodeID(1),
            NodeID(2),
            NodeID(3),
            NodeID(4),
        }

    finally:
        for node_id in range(0, 5):
            node = NodeID(node_id)
            dut[node].stop()



'''Constants that are used throughout the module'''
from enum import Enum
from typing import List, FrozenSet, Tuple
from pathlib import Path

# Copyright (c) 2023, 2024 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.


# We use this to identify the whole implementation. We use it for
# entry point names, and directories that contain extra configuration
# like ~/.ngautonml/.
PACKAGE_NAME = 'ngautonml'

# We ignore this subdirectory name when autoloading catalogs.
CATALOG_IGNORE = 'impl'


class JSONKeys(Enum):
    '''Standard keys in JSON requests for fit and predict'''
    DATA = 'data'


class ColumnType(Enum):
    '''Data type of a column'''
    NUMERIC = 'numeric'
    INTEGRAL = 'integral'
    CATEGORICAL = 'categorical'
    OTHER = 'other'

    @classmethod
    def list(cls) -> List[str]:
        '''Returns a list of all the valid column types.'''
        def getv(cls_elt: 'ColumnType') -> str:
            return getattr(cls_elt, 'value', '')
        return list(map(getv, cls))

    @classmethod
    def from_str(cls, col_type: str) -> 'ColumnType':
        '''Return a ColumnType from a str, raise a ValueError otherwise.'''
        def getnv(cls_elt: 'ColumnType') -> Tuple[str, str]:
            return (getattr(cls_elt, 'name', ''), getattr(cls_elt, 'value', ''))
        col_type = col_type.lower()
        for name, value in list(map(getnv, cls)):
            if value == col_type:
                return cls[name]
        raise ValueError(f'{col_type} is not a valid {cls.__name__}. '
                         f'Must be one of: {cls.list()}')


class Defaults():
    '''Default values for when not set in JSON problem definition'''
    DATA_TYPE = 'tabular'

    # for the image_dir image classification loader, target name does not need
    #   to  be specified by user, so we set a default.
    IMAGE_CLF_TARGET_NAME = 'class'

    # When we save data files, default to csv.
    FILE_TYPE = 'csv'

    # Frequency of time series for forecasting
    FREQUENCY = 'ME'

    # TODO(piggy): If we introduce an ExecutorCatalog, this should be read
    # from the catalog.
    # TODO(merritt): While we don't have an ExecutorCatalog,
    # the values of INSTANTIANTIONS should be of type Instantiations(Constant)
    INSTANTIANTIONS = ['stub_executor_kind', 'simple']

    # Timeout, in seconds, for listeners
    LISTENER_TIMEOUT = 2.0

    # Default loaded_format when input_format does not have a specifically associated default
    LOADED_FORMAT = 'pandas_dataframe'

    # Random seed
    SEED = 1701

    # How much of the dataset should we use for training?
    SPLIT_FRACTION = 0.8


class Filename(Enum):
    '''Standard output filenames.'''
    TRAIN_PREDICTIONS = 'train'
    TEST_PREDICTIONS = 'test'


class FileType(Enum):
    '''Type of file to save output predictions as.'''
    CSV = 'csv'
    ARFF = 'arff'

    @classmethod
    def list(cls) -> List[str]:
        '''Returns a list of all the valid file extensions.'''
        def getv(cls_elt: 'FileType') -> str:
            return getattr(cls_elt, 'value', '')
        return list(map(getv, cls))


class Matcher(Enum):
    '''Ways that hyperparameter overrides can be matched to pipeline steps'''
    # The Designator for the desired bound pipeline. Defaults to the
    # last queried step if no step is otherwise selected.
    DESIGNATOR = 'designator'

    # The name of the desired step(s).
    NAME = 'name'

    # The name of the algorithm associated with the desired step(s).
    ALGORITHM = 'algorithm'

    # A dict of key/value pairs selecting specific tag values from the
    # algorithm associated with the desired step(s).
    TAGS = 'tags'

    def __str__(self) -> str:
        return f'Matcher.{self.name}'


MATCHERS_TO_LOWERCASE = {Matcher.ALGORITHM, Matcher.DESIGNATOR, Matcher.NAME}


class OutputColName(Enum):
    '''Standard column names for output dataframes'''
    GROUND_TRUTH = 'ground truth'
    DESIGNATOR = 'pipeline'


class OutputFolder():
    '''Names of folders created to store output'''
    TRAIN_PREDICTIONS = Path('train_predictions')
    TEST_PREDICTIONS = Path('test_predictions')
    MODELS = Path('models')
    PIPELINES = Path('pipelines')


class ProblemDefControl(Enum):
    '''Standard values used in the JSON problem definition'''
    DISABLE_GRID_SEARCH = 'disable_grid_search'


class ProblemDefKeys(Enum):
    '''Standard top-level keys used in the JSON problem definition.'''
    # Top-level keys
    # TODO(piggy): Long term, we want these to be in ProblemDef or a closely
    #   related file.  They are still here for now until we can solve the problem
    #   of top-level ConfigComponents being able to know their own name without
    #   circular imports.
    # pylint: disable=duplicate-code
    TASK = 'problem_type'
    DATASET = 'dataset'
    METRICS = 'metrics'
    OUTPUT = 'output'
    HYPERPARAMS = 'hyperparams'
    CROSS_VALIDATION = 'cross_validation'
    AGGREGATION = 'aggregation'
    CONTROL = 'control'

    # Dataset keys
    CONFIG = 'config'
    COL_ROLES = 'column_roles'
    TRAIN_PATH = 'train_path'
    TEST_PATH = 'test_path'
    # pylint: disable=duplicate-code
    TRAIN_DIR = 'train_dir'  # used for ImageDirDataLoader
    TEST_DIR = 'test_dir'  # used for ImageDirDataLoader
    # pylint: enable=duplicate-code
    STATIC_EXOGENOUS_PATH = 'static_exogenous_path'  # For forecasting problems only

    # Column roles keys
    POS_LABEL = 'pos_label'
    COL_NAME = 'name'


# TODO(Merritt): this is deprecated, replace with REQUIRED_KEYS and
#   ALLOWED_KEYS classmethod in ConfigComponent
class ProblemDefKeySet():
    '''Standard sets of problem def keys'''

    class DATASET():
        '''Standard sets of keys for the dataset config clause.'''
        REQUIRED: FrozenSet[str] = frozenset([
            ProblemDefKeys.CONFIG.value])
        ALLOWED: FrozenSet[str] = REQUIRED.union(frozenset([
            ProblemDefKeys.COL_ROLES.value,
            ProblemDefKeys.TEST_PATH.value,
            ProblemDefKeys.TRAIN_PATH.value,
            ProblemDefKeys.TRAIN_DIR.value,
            ProblemDefKeys.TEST_DIR.value,
            ProblemDefKeys.STATIC_EXOGENOUS_PATH.value]))

        class ROLES():
            '''Metadata for roles'''
            REQUIRED: FrozenSet[str] = frozenset([
                ProblemDefKeys.COL_NAME.value])
            ALLOWED: FrozenSet[str] = REQUIRED.union(frozenset([
                ProblemDefKeys.POS_LABEL.value
            ]))


class RangeMethod(Enum):
    '''Methods of specifying the range of a hyperparameter for tuning'''
    # We provide exactly the one fixed value in range.
    FIXED = 'fixed'
    # Search each of the values in range.
    LIST = 'list'
    # Search a linear range of values in range: [start, increment, end]
    LINEAR = 'linear'

    def __str__(self) -> str:
        return f'RangeMethod.{self.name}'



'''Tests for neighbor_metadata.py'''

import time

import pytest

from .toy_neighbor_state import ToyNeighborState

from .neighbor_metadata import NeighborMetadata


ANCIENT_DURATION = 100.0  # seconds since the start of the monotonic clock.
SHORT_SLEEP = 0.1  # seconds


def test_neighbor_metadata():
    '''Test the NeighborMetadata class.'''
    metadata = NeighborMetadata()

    # We start out with reasonable defaults.
    assert metadata.current_state is None
    assert metadata.last_state is None
    assert metadata.last_state_sent is None
    assert metadata.cumulative_payload_sent == 0
    # Times have not been updated and are initialized to 0.0.
    assert metadata.time_since_last_state_update > ANCIENT_DURATION
    assert metadata.time_since_last_state_sent > ANCIENT_DURATION

    state = ToyNeighborState(['a', 'b'], 0.0)
    new_state = ToyNeighborState(['a', 'b'], 1.0)

    # Does the current state migrate to the last state?
    metadata.current_state = state
    assert metadata.current_state == state
    assert metadata.last_state is None

    # Does setting the last state send time update the time?
    metadata.last_state_sent = state
    assert metadata.cumulative_payload_sent == state.payload_size
    assert metadata.last_state_sent == state
    duration1 = metadata.time_since_last_state_sent
    assert duration1 == pytest.approx(0.0, abs=0.01)  # BUG: This might be a source of flakiness.
    time.sleep(SHORT_SLEEP)
    duration2 = metadata.time_since_last_state_sent
    assert duration2 > duration1

    # Can we set the current state to None?
    metadata.current_state = None
    assert metadata.last_state == state
    assert metadata.current_state is None
    duration3 = metadata.time_since_last_state_update
    # BUG: This might be a source of flakiness. Set abs higher if this flakes.
    assert duration3 == pytest.approx(0.0, abs=0.01)
    # Can we set the last state sent to None?
    metadata.last_state_sent = None
    assert metadata.last_state_sent is None
    duration4 = metadata.time_since_last_state_sent
    assert duration4 == pytest.approx(0.0, abs=0.01)  # BUG: This might be a source of flakiness.
    assert metadata.cumulative_payload_sent == state.payload_size

    # Can we set the current state to a new state?
    metadata.current_state = state
    metadata.last_state_sent = state
    assert metadata.last_state is None
    assert metadata.current_state == state
    assert metadata.cumulative_payload_sent == 2 * state.payload_size

    # Let's send a different state.
    metadata.current_state = new_state
    metadata.last_state_sent = new_state
    assert metadata.last_state_sent == new_state
    assert metadata.cumulative_payload_sent == 2 * state.payload_size + new_state.payload_size


def test_scratch():
    '''Test the scratchpad.'''
    metadata = NeighborMetadata()
    assert len(metadata.scratch) == 0
    metadata.scratch['a'] = 1
    assert metadata.scratch['a'] == 1



'''Test object to register in a Widget catalog.'''
from typing import Optional

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from ngautonml.catalog.catalog import upcast
from ngautonml.catalog.widgets.impl.widget_catalog import WidgetCatalog, NamedWidget


class Gizmo(NamedWidget):
    '''Test object to register in a Widget catalog.'''
    tags = {
        'some_tag': ['test_dir'],
    }
    frob: Optional[str]
    quux: Optional[str]

    def __init__(self, frob: Optional[str] = None, quux: Optional[str] = None):
        super().__init__('Gizmo')
        self.frob = frob
        self.quux = quux


def register(catalog: WidgetCatalog, *args, **kwargs) -> None:
    '''Register all the objects in this file.'''
    gizmo = Gizmo(*args, **kwargs)
    catalog.register(gizmo, gizmo.name, upcast(gizmo.tags))



'''Test for issue #1 on gitlab (https://gitlab.com/autonlab/ngautonml/-/issues/1)'''
# Copyright (c) 2024 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

# pylint: disable=missing-function-docstring,redefined-outer-name,duplicate-code

import pandas as pd
import pytest

from ..problem_def.problem_def import ProblemDefinition
from ..wrangler.wrangler import Wrangler


@pytest.fixture(scope='session')
def bug1_problem_def() -> ProblemDefinition:
    config = {
        "dataset": {
            "config": "memory",
            "column_roles": {
                "target": {
                    "name": "c"
                }
            },
            "params": {
                "train_data": "train_data",
                "test_data": "test_data"
            }
        },
        "problem_type": {
            "data_type": "tabular",
            "task": "binary_classification"
        },
        "metrics": {
            "roc_auc_score": {},
            "accuracy_score": {}
        },
        "hyperparams": [
            "disable_grid_search"
        ]
    }

    return ProblemDefinition(clause=config)


def test_bug1(bug1_problem_def: ProblemDefinition) -> None:
    train_data = pd.DataFrame(  # noqa: F841 pylint: disable=unused-variable
        {
            'a': range(1, 101),
            'b': range(101, 201),
            'c': [0, 1] * 50
        }
    )
    test_data = pd.DataFrame(  # noqa: F841 pylint: disable=unused-variable
        {
            'a': range(101, 151),
            'b': range(201, 251),
            'c': [0, 1] * 25
        }
    )
    dut = Wrangler(
        problem_definition=bug1_problem_def)

    got = dut.fit_predict_rank()
    assert got.test_results is not None
    pred1 = got.test_results.predictions

    new_data = dut.load_test_dataset()
    assert new_data is not None
    got2 = dut.predict(new_data=new_data)
    pred2 = got2.predictions

    for des in pred1.keys():
        pred1_des = pred1[des]
        pred2_des = pred2[des]
        assert pred1_des is not None
        assert pred2_des is not None
        df1 = pred1_des.predictions
        df2 = pred2_des.predictions
        print(des)
        print(pd.concat([df1, df2], axis=1))
        pd.testing.assert_frame_equal(df1, df2)



'''Tests for the Instantiator class implementations'''

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from .instantiator import InstantiatorStub
from ..generator.bound_pipeline import BoundPipelineStub
from ..generator.designator import Designator

# pylint: disable=missing-function-docstring,duplicate-code
# pylint: disable=missing-class-docstring
# pylint: disable=too-few-public-methods


def test_sunny_day() -> None:
    dut = InstantiatorStub()
    bpl = BoundPipelineStub('bound_pipeline_stub')
    res = dut.instantiate(pipeline=bpl)
    assert res is not None
    assert 'bound_pipeline_stub' == res.name


def test_instantiate_all() -> None:
    dut = InstantiatorStub()
    bpl1 = BoundPipelineStub('bound_pipeline_stub_1')
    bpl2 = BoundPipelineStub('bound_pipeline_stub_2')
    des1 = Designator('des1')
    des2 = Designator('des2')
    res = dut.instantiate_all(pipelines={
        des1: bpl1,
        des2: bpl2})
    assert des1 in res and des2 in res
    assert 'bound_pipeline_stub_1' == res[des1].name
    assert 'bound_pipeline_stub_2' == res[des2].name



'''Base class for Wranglers that can fit, predict and rank sets of pipelines.

The Wrangler does the automl pieces, this base class does the rest.
'''

# Copyright (c) 2024 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

# pylint: disable=too-many-instance-attributes,too-many-arguments

import abc
from pathlib import Path
from typing import Any, Dict, Iterable, List, Optional, Type, Union

from ..aggregators.impl.aggregator_auto import AggregatorCatalogAuto
from ..aggregators.impl.aggregator_catalog import AggregatorCatalog
from ..algorithms.impl.algorithm import AlgorithmCatalog
from ..algorithms.impl.algorithm_auto import AlgorithmCatalogAuto
from ..catalog.catalog import Catalog
from ..communicators.impl.communicator_auto import CommunicatorCatalogAuto
from ..communicators.impl.communicator_catalog import CommunicatorCatalog
from ..config_components.dataset_config import DatasetConfig
from ..cross_validators.impl.cross_validator import CrossValidatorCatalog
from ..cross_validators.impl.cross_validator_auto import \
    CrossValidatorCatalogAuto
from ..data_loaders.impl.data_loader import DataLoader
from ..data_loaders.impl.data_loader_auto import DataLoaderCatalogAuto
from ..data_loaders.impl.data_loader_catalog import DataLoaderCatalog
from ..discoverers.impl.discoverer_auto import DiscovererCatalogAuto
from ..discoverers.impl.discoverer_catalog import DiscovererCatalog
from ..executor.executor import Executor
from ..executor.simple.simple_executor import SimpleExecutor
from ..generator.bound_pipeline import BoundPipeline
from ..generator.designator import Designator
from ..generator.generator import Generator, GeneratorImpl
from ..instantiator.executable_pipeline import (ExecutablePipeline,
                                                PipelineResults)
from ..instantiator.instantiator_factory import InstantiatorFactory
from ..metrics.impl.metric import Metric, MetricInvalidDatasetError
from ..metrics.impl.metric_catalog import MetricCatalog
from ..metrics.impl.metric_auto import MetricCatalogAuto
from ..pipeline_loaders.impl.pipeline_loader_auto import PipelineLoaderCatalogAuto
from ..pipeline_loaders.impl.pipeline_loader_catalog import PipelineLoaderCatalog
from ..problem_def.problem_def import ProblemDefinition
from ..ranker.ranker import Ranker, Rankings
from ..ranker.ranker_impl import RankerImpl
from ..searcher.searcher import Searcher, SearcherImpl
from ..splitters.impl.splitter import SplitterCatalog
from ..splitters.impl.splitter_auto import SplitterCatalogAuto
from ..templates.impl.pipeline_template import PipelineTemplate
from ..templates.impl.template import TemplateCatalog
from ..templates.impl.template_auto import TemplateCatalogAuto

from .dataset import Dataset
from .logger import Level, Logger
from .saver import Saver

log = Logger(__file__, level=Level.WARN, to_stdout=True).logger()


class Error(Exception):
    '''Base error for Wrangler.'''


class WranglerFailure(Error):
    '''Wrangler can not procede.'''


class ReassignmentError(Error):
    '''An attempt was made to assign a value more than once.'''


class WranglerBase(metaclass=abc.ABCMeta):  # pylint: disable=too-many-public-methods
    '''Base class for Wranglers that can fit, predict and rank sets of pipelines.

    The Wrangler does the automl pieces, this base class does the rest.
    '''
    _bound_pipelines: Optional[Dict[Designator, BoundPipeline]] = None
    # These are all pipelines identified during the first run.
    _all_executable_pipelines: Optional[Dict[Designator, ExecutablePipeline]] = None
    # This is a subset of _all_pipelines which will be the default
    # for subsequent fit, predict, and rank methods.
    _current_executable_pipelines: Optional[Dict[Designator, ExecutablePipeline]] = None

    def __init__(
        self,
        problem_definition: ProblemDefinition,
        aggregator_catalog: Optional[Type[AggregatorCatalog]] = None,
        algorithm_catalog: Optional[Type[AlgorithmCatalog]] = None,
        communicator_catalog: Optional[Type[CommunicatorCatalog]] = None,
        dataloader_catalog: Optional[Type[DataLoaderCatalog]] = None,
        discoverer_catalog: Optional[Type[DiscovererCatalog]] = None,
        executor: Optional[Type[Executor]] = None,
        generator: Optional[Type[Generator]] = None,
        instantiator_factory: Optional[Type[InstantiatorFactory]] = None,
        metric_catalog: Optional[Type[MetricCatalog]] = None,
        pipeline_loader_catalog: Optional[Type[PipelineLoaderCatalog]] = None,
        ranker: Optional[Type[Ranker]] = None,
        saver: Optional[Type[Saver]] = None,
        searcher: Optional[Type[Searcher]] = None,
        splitter_catalog: Optional[Type[SplitterCatalog]] = None,
        template_catalog: Optional[Type[TemplateCatalog]] = None,
        validator_catalog: Optional[Type[CrossValidatorCatalog]] = None
    ):
        # TODO(Merritt): initialize the logger
        # instantiate default components only if input is None
        self._pd = problem_definition

        self._metric_catalog = (metric_catalog or MetricCatalogAuto)()
        self._communicator_catalog = (communicator_catalog or CommunicatorCatalogAuto)()
        self._discoverer_catalog = (discoverer_catalog or DiscovererCatalogAuto)()
        self._algorithm_catalog = (algorithm_catalog or AlgorithmCatalogAuto)(
            communicator_catalog=self._communicator_catalog,
            discoverer_catalog=self._discoverer_catalog
        )
        self._generator = (generator or GeneratorImpl)(
            algorithm_catalog=self._algorithm_catalog,
            problem_definition=self._pd)
        self._template_catalog = (template_catalog or TemplateCatalogAuto)(
            algorithm_catalog=self._algorithm_catalog,
            generator=self._generator
        )
        self._pipeline_loader_catalog = (pipeline_loader_catalog or PipelineLoaderCatalogAuto)(
            algorithm_catalog=self._algorithm_catalog,
            template_catalog=self._template_catalog)
        self._searcher = (searcher or SearcherImpl)(hyperparams=self._pd.hyperparams)
        self._executor = (executor or SimpleExecutor)()
        self._ranker = (ranker or RankerImpl)()
        self._splitter_catalog = (splitter_catalog or SplitterCatalogAuto)(
            cv_config=self._pd.cross_validation_config)
        self._dataloader_catalog = (dataloader_catalog or DataLoaderCatalogAuto)()
        self._dataloader: Optional[DataLoader] = None
        self._validator_catalog = (validator_catalog or CrossValidatorCatalogAuto)()
        self._aggregator_catalog = (aggregator_catalog or AggregatorCatalogAuto)()
        self._saver = None
        # If there is no output path, we'll output no files.
        if self._pd.output.path is not None:
            self._saver = (saver or Saver)(self._pd.output)
        self._instantiator_factory = (
            instantiator_factory or InstantiatorFactory)(
                saver=self._saver)

    @property
    def aggregator_catalog(self) -> AggregatorCatalog:
        '''Query and register aggregators'''
        return self._aggregator_catalog

    @property
    def algorithm_catalog(self) -> AlgorithmCatalog:
        '''Query and register algorithms'''
        return self._algorithm_catalog

    @property
    def bound_pipelines(self) -> Dict[Designator, BoundPipeline]:
        '''The set of pipelines that have been bound.'''
        assert self._bound_pipelines is not None, (
            'BUG: bound_pipelines accessed before it is set.'
        )
        return self._bound_pipelines

    @property
    def communicator_catalog(self) -> CommunicatorCatalog:
        '''Query and register communicators'''
        return self._communicator_catalog

    @property
    def dataloader_catalog(self) -> DataLoaderCatalog:
        '''Load data based on input and loaded types.'''
        return self._dataloader_catalog

    @property
    def discoverer_catalog(self) -> DiscovererCatalog:
        '''Query and register discoverers.'''
        return self._discoverer_catalog

    @property
    def metric_catalog(self) -> Catalog[Metric]:
        '''Query and register metrics.'''
        return self._metric_catalog

    @property
    def metrics(self) -> Dict[str, Metric]:
        '''The metrics specified in the problem definition.'''
        return self._metric_catalog.lookup_metrics(self._pd)

    @property
    def pipeline_loader_catalog(self) -> PipelineLoaderCatalog:
        '''Query and register pipeline loaders.'''
        return self._pipeline_loader_catalog

    @property
    def splitter_catalog(self) -> SplitterCatalog:
        '''Query and register splitters.'''
        return self._splitter_catalog

    @property
    def template_catalog(self) -> TemplateCatalog:
        '''Query and register templates.'''
        return self._template_catalog

    @property
    def validator_catalog(self) -> CrossValidatorCatalog:
        '''Query and register cross validators.'''
        return self._validator_catalog

    @property
    def generator(self) -> Generator:
        '''Generates bound pipelines from templates.'''
        return self._generator

    @property
    def ranker(self) -> Ranker:
        '''Ranks the results of running pipelines.'''
        return self._ranker

    @property
    def all_trained(self) -> bool:
        '''Are all the current pipelines trained?'''
        if self._current_executable_pipelines is None:
            return True
        return all(pipeline.trained
                   for pipeline in self._current_executable_pipelines.values())

    @property
    def locked(self) -> bool:
        '''Does any current pipeline contain a model that is currently locked?'''
        if self._current_executable_pipelines is None:
            return False
        return any(pipe.locked
                   for pipe in self._current_executable_pipelines.values())

    def converged(self) -> bool:
        '''Are all the models in all the current pipelines converged?'''

        if self._current_executable_pipelines is None:
            return False
        return all(pipe.converged()
                   for pipe in self._current_executable_pipelines.values())

    def _find_data_loader(self) -> DataLoader:
        if self._dataloader is not None:
            return self._dataloader
        dataset_config = self._pd.get_conf(self._pd.Keys.DATASET.value)  # type: ignore[attr-defined] # pylint: disable=line-too-long
        assert isinstance(dataset_config, DatasetConfig)
        self._dataloader = self._dataloader_catalog.construct_instance(dataset_config)
        return self._dataloader

    def load_train_dataset(self) -> Optional[Dataset]:
        '''Load this wrangler's train data.'''
        return self._find_data_loader().load_train()

    def load_test_dataset(self) -> Optional[Dataset]:
        '''Load this wrangler's test data, without the target column.'''
        return self._find_data_loader().load_test()

    def load_ground_truth_dataset(self) -> Optional[Dataset]:
        '''Load the target column for this wrangler's test data.'''
        return self._find_data_loader().load_ground_truth()

    def dataset(self, data: Any, **kwargs) -> Dataset:
        '''Load in-memory data into a Dataset object, stored at key 'key'.'''
        return self._find_data_loader().dataset(data=data, **kwargs)

    def save(self, train_results: PipelineResults) -> Dict[Designator, Path]:
        '''Save all the models and all the pipelines.'''
        assert self._saver is not None, 'BUG: Only call save if there is a saver.'
        retval: Dict[Designator, Path] = {}
        model_paths = self._saver.save_models(train_results.executable_pipelines)
        for kind in self._pd.output.instantiations:
            instantiatior = self._instantiator_factory.build(kind)
            for des, result in train_results.items():
                retval[des] = instantiatior.save(result.bound_pipeline, model_paths)
        return retval

    def _build_pipelines(self) -> Dict[Designator, BoundPipeline]:
        '''Build a set of pipelines to use for machine learning'''
        templates = self.lookup_templates()
        if len(templates) == 0:
            raise WranglerFailure(f'found no templates for {self._pd.task}')
        log.info('Found %d templates', len(templates))
        print(f'Found {len(templates)} templates')

        gen_result = self._generator.generate_all(templates)
        log.info("Generated %d bound pipelines", len(gen_result))
        print(f"Generated {len(gen_result)} bound pipelines")

        retval = self._searcher.bind_all(gen_result)
        log.info('Hyperparam searcher produced %d bound pipelines', len(retval))
        print(f'Hyperparam searcher produced {len(retval)} bound pipelines')

        return retval

    def fit(self,
            dataset: Optional[Dataset],
            pipelines: Optional[Dict[Designator, ExecutablePipeline]] = None
            ) -> Dict[Designator, ExecutablePipeline]:
        '''Fit the given pipelines with a new dataset.

        If no pipelines are specified, use the current set.
        '''
        if pipelines is None:
            pipelines = self._current_executable_pipelines

        assert pipelines is not None
        self._executor.fit(
            dataset=dataset,
            pipelines=pipelines)

        return pipelines

    def predict(self,
                new_data: Optional[Dataset] = None,
                trained_pipelines: Optional[
                    Dict[Designator, ExecutablePipeline]] = None) -> PipelineResults:
        '''Predict on new test data that was not supplied in the problem definition.'''
        if trained_pipelines is None:
            trained_pipelines = self._current_executable_pipelines
        assert trained_pipelines is not None, (
            'BUG: predict called with None current pipelines.'
        )
        pipeline_errors: List[str] = []
        for des, pipe in trained_pipelines.items():
            if pipe.kind != self._executor.kind:
                pipeline_errors.append(
                    f'Pipeline {des} has incorrect ExecutorKind {pipe.kind}; '
                    f'expected {self._executor.kind}. \n')
            elif not pipe.trained:
                pipeline_errors.append(
                    f'Pipeline {des} is untrained. \n')

        if len(pipeline_errors) > 0:
            raise WranglerFailure(
                'Bad pipeline(s) provided to predict(): \n'
                f'{"".join(pipeline_errors)}')

        if new_data is None:
            new_data = self.load_test_dataset()
        if new_data is None:
            raise WranglerFailure('No data provided to predict() and no test data specified '
                                  'in the problem definition.  Nothing to predict.')
        results = self._executor.predict(dataset=new_data, pipelines=trained_pipelines)
        return PipelineResults(results)

    def set_current(self, *args: str) -> None:
        '''Set the current set of pipelines to work with by name.

        If no arguments are given, reset to the full set of pipelines
        identified in the last fit_predict_rank.
        '''
        if len(args) == 0:
            self._current_executable_pipelines = self._all_executable_pipelines
            return
        if self._all_executable_pipelines is None:
            self._current_executable_pipelines = None
            return
        self._current_executable_pipelines: Dict[Designator, ExecutablePipeline] = {
            Designator(des): self._all_executable_pipelines[Designator(des)]
            for des in args
        }

    def roc_curve(self,
                  results: Union[PipelineResults, Dict[Designator, Optional[Dataset]]],
                  ground_truth: Dataset,
                  metric: Union[str, Metric] = 'roc_curve') -> Dict[
                      Designator, Union[Dataset, Exception, None]]:
        '''Compute the ROC curve for a set of pipeline results.

        Returns:
            A dictionary of results, where the key is the designator of the pipeline
            and the value is the ROC curve. Eash Dataset contains a DataFrame with
            columns 'fpr', 'tpr', and 'thresholds'.

            The value will be None if the input dataset is None.

            If an expected type of Exception is raised while calculating,
            it will return it in the place of a Dataset.
        '''
        if isinstance(results, PipelineResults):
            results = results.predictions

        if isinstance(metric, str):
            metric = self._metric_catalog.lookup_by_name(metric)

        retval: Dict[Designator, Union[Dataset, Exception, None]] = {}

        for des, pred in results.items():
            if pred is None:
                retval[des] = pred
            else:
                try:
                    retval[des] = metric.calculate_roc_curve(pred=pred, ground_truth=ground_truth)
                except (NotImplementedError, MetricInvalidDatasetError) as err:
                    retval[des] = err

        return retval

    def rank(self,
             results: PipelineResults,
             metrics: Optional[Iterable[Union[str, Metric]]] = None,
             ground_truth: Optional[Dataset] = None) -> Rankings:
        '''Rank a set of pipeline results.

        If no metrics are provided, uses the ones from the problem definition

        If no ground truth is provided, attempts to infer it using the SplitDataset
        associated with an arbitrary PipelineResult.
        '''
        metrics_for_ranker: Dict[str, Metric] = {}
        if metrics is None:
            metrics_for_ranker = self._metric_catalog.lookup_metrics(self._pd)
        else:
            for m in metrics:
                if isinstance(m, Metric):
                    metrics_for_ranker[m.name] = m
                    continue

                if isinstance(m, str):
                    metrics_for_ranker[m] = self._metric_catalog.lookup_by_name(m)
                    continue

                raise ValueError(
                    'Element of metrics dict is neither Metric nor str.  '
                    f'Instead found {m} of type {type(m)}.'
                )
        if ground_truth is None:
            ground_truth = self._find_data_loader().load_ground_truth()
        if ground_truth is None:
            ground_truth = results.infer_ground_truth()

        return self.ranker.rank(results=results,
                                metrics=metrics_for_ranker,
                                ground_truth=ground_truth)

    @abc.abstractmethod
    def lookup_templates(self) -> Dict[str, PipelineTemplate]:
        '''Look up the templates that match the problem definition.'''



'''Discoverer that uses dynamic configuration.'''

# Copyright (c) 2024 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from queue import Queue
from typing import Dict, List

from .impl.discoverer import Discoverer
from .impl.discoverer_catalog import DiscovererCatalog


class DynamicDiscoverer(Discoverer):
    '''Discoverer that dynamically discovers neighbors.'''
    name = 'dynamic'
    tags: Dict[str, List[str]] = {}

    def start(self, queue: Queue):
        self._communicator.start(
            queue=queue
        )

    def laplacian(self) -> List[List[int]]:
        raise NotImplementedError('DynamicDiscoverer does not support laplacian')


def register(catalog: DiscovererCatalog) -> None:
    '''Register all objects in this file.'''
    catalog.register(DynamicDiscoverer)



'''Test the DeciderCatalog'''

# Copyright (c) 2024 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from ..stub_decider import DeciderStub
from .decider_catalog import DeciderCatalogStub

# pylint: disable=missing-function-docstring,duplicate-code


def test_register() -> None:
    dut = DeciderCatalogStub()
    assert 'stub_decider' == dut.register(obj=DeciderStub)


def test_register_name() -> None:
    dut = DeciderCatalogStub()
    stub_decider = DeciderStub
    assert 'somenameddecider' == dut.register(
        obj=stub_decider, name='SomeNamedDecider')


def test_lookup_by_name() -> None:
    dut = DeciderCatalogStub()
    some_decider = DeciderStub
    dut.register(obj=some_decider, name='SomeNamedDecider')
    assert dut.lookup_by_name('SomeNamedDecider') == some_decider


def test_lookup_one_by_tag() -> None:
    dut = DeciderCatalogStub()
    dut.register(obj=DeciderStub, name='NamedStubDecider', tags={'k1': ['v1']})
    assert len(dut.lookup_by_tag_and(k1='v1')) == 1


def test_lookup_by_tag_fail() -> None:
    dut = DeciderCatalogStub()
    dut.register(obj=DeciderStub, name='NamedStubDecider', tags={'k1': ['v1']})
    assert len(dut.lookup_by_tag_and(k1='different_value')) == 0


def test_lookup_multi_by_tag() -> None:
    dut = DeciderCatalogStub()
    stub_decider = DeciderStub
    dut.register(obj=stub_decider, name="Name1", tags={'k1': ['v1']})
    stub_decider2 = DeciderStub
    dut.register(obj=stub_decider2, name="Name2", tags={'k1': ['v1']})
    stub_decider3 = DeciderStub
    dut.register(obj=stub_decider3, name="Name3", tags={'k1': ['different_value']})
    assert len(dut.lookup_by_tag_and(k1='v1')) == 2



'''Tests for pipeline_loader.py'''

# pylint: disable=missing-function-docstring, missing-class-docstring,duplicate-code

from ...generator.bound_pipeline import BoundPipeline

from .pipeline_loader import PipelineLoader


class FakePipelineLoader(PipelineLoader):
    def _load(self, *unused_args, **unused_kwargs) -> BoundPipeline:
        return BoundPipeline(name='fake_pipeline_loader')


def test_load() -> None:
    dut = FakePipelineLoader()
    got = dut.load(name='fake_pipeline_loader')
    assert isinstance(got, BoundPipeline)



'''Tests for static_discoverer.py'''

from ..communicators.stub_communicator import CommunicatorStub
from ..config_components.distributed_config import DistributedConfig
from ..neighbor_manager.node_id import NodeID

from .static_discoverer import StaticDiscoverer


def test_lapacian():
    '''Test the lapacian function'''
    config = DistributedConfig(clause={
        'discoverer': {
            'name': 'static',
            'static': {
                'adjacency': {
                    '6': [4],
                    '5': [1, 2, 4],
                    '4': [3, 5, 6],
                    '3': [2, 4],
                    '2': [1, 3, 5],
                    '1': [2, 5],
                },
            },
        },
        'communicator': {'name': 'stub'},
        'my_id': 1
    })

    # See https://en.wikipedia.org/wiki/Laplacian_matrix https://tinyurl.com/mr2vu3hd
    want = [
        [ 2, -1,  0,  0, -1,  0],   # noqa: E241, E201
        [-1,  3, -1,  0, -1,  0],   # noqa: E241, E201
        [ 0, -1,  2, -1,  0,  0],   # noqa: E241, E201
        [ 0,  0, -1,  3, -1, -1],   # noqa: E241, E201
        [-1, -1,  0, -1,  3,  0],   # noqa: E241, E201
        [ 0,  0,  0, -1,  0,  1]    # noqa: E241, E201
    ]

    communicator = CommunicatorStub(my_id=NodeID(1))
    discoverer = StaticDiscoverer(config=config, communicator=communicator)

    got = discoverer.laplacian()
    assert want == got
